#coding:utf-8
from typing import Optional as Opt
from sys import stderr
import json
from logging import Logger, Formatter, StreamHandler, Manager, addLevelName, getLevelName, root
from collections import Counter
from .defaults import PROGRESS, ERROR, PROGRESS_LEVEL, METALOG, METALOG_LEVEL
from .timing import TimedTaskContext
from .helpers import *

logger_method_names = ("debug", "info", "warning", "error", "critical")


# a callable logger that can stand in for the print() function in any codebase
# at the debug level
class CallableLogger(Logger):
    def __call__(self, *args, **kwargs):
        self.debug(*args, **kwargs)


# this is a preconfigured logger that just prints to stdout - all of its
# logging methods behave just like `print`, and it's callable - so you can put
# 'import logging.printLogger as print' at the top of any python3 file and all
# of your print statements will log to stderr with the same contents as the
# original print statements.
printLogger = CallableLogger("print", level="DEBUG")
basicFormatter = Formatter("%(message)s")
printHandler = StreamHandler(stderr)
printHandler.setFormatter(basicFormatter)
printLogger.addHandler(printHandler)
printLogger.parent = None


class CountingLogger(CallableLogger):
    """
    This logger subclass allows validation of log file parses by counting the number of
    times that it has logged at each level, how many stack traces it has logged, and how
    many multiline messages it has logged (which are potentially ambiguous to parse)
    """
    addLevelName(METALOG, METALOG_LEVEL)
    manager = Manager(root)

    def __init__(self, *args, **kwargs):
        self.multiline = 0
        self.stacktraces = 0
        self.total = 0
        self.levelcounts = Counter()
        super().__init__(*args, **kwargs)

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False, stats=False):
        levelname = getLevelName(level)

        if not stats:
            super()._log(level, msg, args, exc_info=exc_info, extra=extra, stack_info=stack_info)
            if exc_info:
                self.stacktraces += 1
            if "\n" in msg:
                self.multiline += 1

        if level >= self.level:
            self.levelcounts.update([levelname])

        if stats:
            msg = json.dumps(self.stats, indent=None)
            if "\n" in msg:
                self.multiline += 1
                stats_ = json.loads(msg)
                stats_['multiline'] += 1
                msg = json.dumps(self.stats, indent=None)
            super()._log(level, msg, args, exc_info=exc_info, extra=extra, stack_info=stack_info)

        self.total += 1

    @property
    def stats(self):
        return dict(stacktraces=self.stacktraces, multiline=self.multiline,
                    levelcounts=self.levelcounts)

    def _report_stats(self):
        self._log(METALOG, None, None, stats=True)

    def __del__(self):
        # log the final message count stats on program exit
        self._report_stats()


class ProgressLogger(CallableLogger):
    """
    This logger subclass allows for simple and consistent logging of progress on user-defined
    jobs:
    >>> logger = ProgressLogger("fooLog", job_level="INFO")
    >>> with logger.task("count_to_10", time_units='s') as task:
    >>>     for i in range(10):
    >>>         task.report_progress(1)
    >>>         # verbose output happens here
    """
    addLevelName(PROGRESS, PROGRESS_LEVEL)
    manager = Manager(root)

    def __init__(self, *args, job_level=PROGRESS_LEVEL,
                 **kwargs):
        self.job_level = validate_log_level(job_level)
        super().__init__(*args, **kwargs)

    def task(self, job_name: str,
             total_tasks: Opt[int]=None,
             task_units: Opt[str]=None,
             level=PROGRESS,
             error_level=ERROR,
             time_units: str='s'):
        level = level or self.job_level
        return TimedTaskContext(job_name, total_tasks=total_tasks,
                                task_units=task_units, time_units=time_units,
                                logger_or_print_func=self, level=level, error_level=error_level)


class SwissArmyLogger(ProgressLogger, CountingLogger):
    """
    One logger to do all the things
    """
    manager = Manager(root)

    def __init__(self, *args, **kwargs):
        ProgressLogger.__init__(self, *args, **kwargs)
        CountingLogger.__init__(self, *args, **kwargs)


for cls in (CountingLogger, ProgressLogger, SwissArmyLogger):
    cls.manager.loggerClass = cls



# coding:utf-8
import typing
from argparse import ZERO_OR_MORE
import decimal
import fractions
from bourbaki.introspection.types import NonStrCollection, is_named_tuple_class, get_named_tuple_arg_types
from bourbaki.introspection.generic_dispatch import GenericTypeLevelSingleDispatch, UnknownSignature
from .utils import maybe_map
from .exceptions import CLIIOUndefined


NoneType = type(None)

NestedCollectionTypes = (typing.Collection[NonStrCollection],
                         typing.Tuple[NonStrCollection, ...],
                         typing.Mapping[NonStrCollection, typing.Any],
                         typing.Mapping[typing.Any, NonStrCollection])


class AmbiguousUnionNargs(CLIIOUndefined):
    def __str__(self):
        return ("Types in union {} imply an ambiguous number of command line args"
                .format(self.type_))


class NestedCollectionsCLIArgError(CLIIOUndefined):
    def __str__(self):
        return ("Some type parameters sequence type {} require more than one command line arg; can't parse "
                "unambiguously".format(self.type_))


def check_union_nargs(*types):
    types = [t for t in types if t not in (NoneType, None)]
    all_nargs = tuple(maybe_map(cli_nargs, types, (UnknownSignature, CLIIOUndefined)))
    if len(set(all_nargs)) > 1:
        raise AmbiguousUnionNargs((typing.Union, *types))
    if len(all_nargs) == 0:
        raise CLIIOUndefined((typing.Union, *types))
    return all_nargs


def check_tuple_nargs(t, *types):
    all_nargs = tuple(cli_nargs(t) for t in types if t is not Ellipsis)
    if any(a not in (1, None) for a in all_nargs):
        raise NestedCollectionsCLIArgError((t, *types))
    return all_nargs


# nargs for argparse.ArgumentParser

cli_nargs = GenericTypeLevelSingleDispatch("cli_nargs", isolated_bases=[typing.Union])

cli_nargs.register_all(decimal.Decimal, fractions.Fraction, as_const=True)(None)


@cli_nargs.register(typing.Any)
def default_nargs(*args, **kwargs):
    # single string arg unless otherwise overridden below
    return None


@cli_nargs.register(NonStrCollection)
def seq_nargs(*types):
    # all collections other than str
    return ZERO_OR_MORE


@cli_nargs.register_all(NestedCollectionTypes)
def nested_collections_cli_error(t, *args):
    # collections of collections
    raise NestedCollectionsCLIArgError((t, *args))


@cli_nargs.register(typing.Tuple)
def tuple_nargs(t, *types):
    if not types and is_named_tuple_class(t):
        types = get_named_tuple_arg_types(t)
    elif not types or types[-1] is Ellipsis:
        _ = check_tuple_nargs(t, *types)
        return ZERO_OR_MORE

    _ = check_tuple_nargs(t, *types)
    return len(types)


@cli_nargs.register(typing.Union)
def union_nargs(u, *types):
    all_nargs = check_union_nargs(*types)
    return next(iter(all_nargs))



# !/usr/bin/env python
"""
Usage:
    compgen_python_classpaths.py IMPORT_PREFIX [legal_prefix1 [legal_prefix2 [...]]]
    compgen_python_classpaths.py IMPORT_PREFIX --class [legal_prefix1 [legal_prefix2 [...]]]
    compgen_python_classpaths.py IMPORT_PREFIX --callable [legal_prefix1 [legal_prefix2 [...]]]
    compgen_python_classpaths.py IMPORT_PREFIX --subclass [classpath1 [classpath2 [...]]]
    compgen_python_classpaths.py IMPORT_PREFIX --instance [classpath1 [classpath2 [...]]]
"""

import os
import sys
from typing import Union
from warnings import warn
Module = type(sys)
from itertools import chain

COMPLETION_DEBUG_ENV_VAR = "BOURBAKI_COMPLETION_DEBUG"

DEBUG = os.environ.get(COMPLETION_DEBUG_ENV_VAR, "").lower().strip()
if DEBUG == "true" or (DEBUG.isdigit() and DEBUG != "0"):
    warn("Found environment variable {}={}; verbose output will be generated for {}"
         .format(COMPLETION_DEBUG_ENV_VAR, DEBUG, __file__))
    DEBUG = True if not DEBUG.isdigit() else int(DEBUG)
else:
    DEBUG = False

CLASS_FLAG = '--class'
CALLABLE_FLAG = '--callable'
MODULE_FLAG = '--module'
INSTANCE_FLAG = '--instance'
SUBCLASS_FLAG = '--subclass'


# CLI dispatches to one of these functions

def complete_objpath(prefix: str= '', *legal_prefixes: str):
    _debug("Completing paths for all objects")
    yield from _complete_objpath(prefix, legal_prefixes=legal_prefixes)


def complete_callable_path(prefix: str='', *legal_prefixes: str):
    _debug("Completing paths for callables")
    yield from _complete_objpath(prefix, typecheck=callable_, legal_prefixes=legal_prefixes)


def complete_classpath(prefix: str='', *legal_prefixes: str):
    _debug("Completing paths for classes")
    yield from _complete_objpath(prefix, cls=type, typecheck=isinstance, legal_prefixes=legal_prefixes)


def complete_module_path(prefix: str='', *legal_prefixes: str):
    _debug("Completing paths for modules")
    yield from _complete_objpath(prefix, include_attrs=False, legal_prefixes=legal_prefixes, include_builtins=False)


def complete_instance_path(prefix: str='', *legal_classes: str):
    _debug("Completing paths for instances of : {cls}", cls=legal_classes)
    if legal_classes:
        cls = _materialize_classes(*legal_classes)
        typecheck = isinstance
    else:
        cls = typecheck = None

    yield from _complete_objpath(prefix, cls=cls, typecheck=typecheck)


def complete_subclasspath(prefix: str= '', *legal_superclasses: str):
    _debug("Completing paths for subclasses of : {cls}", cls=legal_superclasses)
    if legal_superclasses:
        cls = _materialize_classes(*legal_superclasses)
        typecheck = issubclass_
    else:
        cls, typecheck = type, isinstance

    yield from _complete_objpath(prefix, cls=cls, typecheck=typecheck)


# The above dispatch to this main function

def _complete_objpath(prefix: str, cls=None, typecheck=None, legal_prefixes=None,
                      include_attrs=True, include_builtins=True):
    from bourbaki.introspection.imports import import_object
    import builtins

    if legal_prefixes:
        candidate_prefixes = [p for p in legal_prefixes if p.startswith(prefix)]
        if not candidate_prefixes:
            return None, ()
    else:
        candidate_prefixes = [prefix]

    def inner(prefix):
        parts = prefix.split('.')
        if len(parts) > 1:
            modname, suffix = '.'.join(parts[:-1]), parts[-1]
            names = get_all_module_names(prefix, suffix)
            try:
                mod = import_object(modname)
            except ImportError:
                mod = None
            else:
                names = chain(names, get_submodule_names(mod, suffix, modname=modname))
                if include_attrs:
                    names = chain(names, get_attr_names(mod, suffix, cls, typecheck, modname=modname))
        else:
            mod = None
            names = get_all_module_names(prefix)
            if include_builtins:
                names = chain(names, get_attr_names(builtins, prefix, cls, typecheck))
        return names

    return chain.from_iterable(map(inner, candidate_prefixes))


# Helpers

def _debug(msg: str, **kw):
    if DEBUG:
        OKGREEN = '\033[92m'
        WARNING = '\033[93m'
        FAIL = '\033[91m'
        ENDC = '\033[0m'
        print("\033[92m{}: {}\033[0m".format(os.path.basename(__file__), msg.format(**kw)),
              file=sys.stderr)


def callable_(o, cls=None):
    return callable(o)


def issubclass_(o, cls):
    # delayed import for speed if not needed; this only gets called at most once per interpreter launch
    from bourbaki.introspection.types import issubclass_generic, typetypes
    return isinstance(o, (type, *typetypes)) and issubclass_generic(o, cls)


def _materialize_classes(*legal_superclasses: str):
    if not legal_superclasses:
        clss = None
    else:
        from bourbaki.introspection.imports import import_type
        from bourbaki.introspection.types import typetypes
        clss = []
        for name in legal_superclasses:
            try:
                cls = import_type(name)
            except (ImportError, AttributeError):
                pass
            else:
                if isinstance(cls, (type, *typetypes)):
                    clss.append(cls)
        clss = tuple(clss)

    _debug("Imported classes: {clss}", clss=clss)
    if not clss:
        return object
    return Union[clss]


def get_attr_names(mod, attr_prefix='', cls=None, typecheck=None, modname=None):
    _debug("Completing attributes on module/object {mod}", mod=mod)

    fmt_name = None if modname is None else (modname + ".{}").format

    attrs = dir(mod)

    if attr_prefix:
        _debug("Filtering to attributes with prefix {prefix}", prefix=prefix)
        attrs = (a for a in attrs if a.startswith(attr_prefix))

    if typecheck is not None:
        _debug("Filtering to attributes by type using {func}({cls})",
               func=typecheck, cls=cls)
        attrs = (a for a in attrs if typecheck(getattr(mod, a, None), cls))

    yield from (map(fmt_name, attrs) if fmt_name else attrs)


def get_submodule_names(mod, suffix='', modname=None):
    file_ = getattr(mod, "__file__", None)

    if modname is None:
        modname = getattr(mod, "__name__", '')

    if file_ and os.path.basename(file_) == '__init__.py':
        dirs = getattr(mod, "__path__", [os.path.dirname(file_)])
        for dir_ in dirs:
            _debug("Completing submodules of {modname} in {dir_}", modname=modname, dir_=dir_)
            for name in get_module_names(dir_, suffix):
                yield '.'.join([modname, name])


def get_module_names(dir_, submod_prefix=None):
    # _debug("Searching modules in {dir_}", dir_=dir_)
    for name in os.listdir(dir_):
        if name == '__init__.py':
            continue
        if submod_prefix and not name.startswith(submod_prefix):
            continue

        prefix, ext = os.path.splitext(name)
        if ext == '.py':
            yield prefix
        elif ext == '.pth':
            yield prefix.split('-')[0]
        elif ext == '.so':
            yield prefix.split('.')[0]
        elif ext == '.egg-link':
            yield prefix.replace('-', '_')
        elif ext not in ('.dist-info', '.egg-info'):
            d = os.path.join(dir_, name)
            if os.path.isdir(d) and os.path.isfile(os.path.join(d, '__init__.py')):
                yield name


def get_all_module_names(prefix='', submod_prefix=None):
    _debug("Listing all known modules from sys.path{prefix_note}",
           prefix_note='' if not prefix else " with prefix '%s'" % prefix)
    dirs = [d for d in sys.path if os.path.isdir(d)]

    names = (name for dir_ in dirs for name in get_module_names(dir_, submod_prefix))
    names = chain(names, ('builtins',))
    if prefix:
        names = (name for name in names if name.startswith(prefix))

    yield from uniq(names)


def uniq(iter_):
    visited = set()
    for _ in iter_:
        if _ in visited:
            continue
        visited.add(_)
        yield _


if __name__ == "__main__":
    argv = sys.argv[1:]
    if not argv or argv[0] in ('-h', '--help'):
        print(__doc__)
        exit(0)

    if os.getenv("APPUTILS_COMPLETION_DEBUG", "false") == "true":
        DEBUG = True

    dispatch = {
        CLASS_FLAG: complete_classpath,
        SUBCLASS_FLAG: complete_subclasspath,
        INSTANCE_FLAG: complete_instance_path,
        MODULE_FLAG: complete_module_path,
        CALLABLE_FLAG: complete_callable_path,
    }

    prefix, args = argv[0], argv[1:]

    if not args:
        args = ['']
        complete = complete_objpath
    elif args[0] in dispatch:
        complete = dispatch[args[0]]
        args = args[1:]
    else:
        complete = complete_objpath

    for name in complete(prefix, *args):
        print(name, file=sys.stdout)



# coding:utf-8
from argparse import ONE_OR_MORE, OPTIONAL, ZERO_OR_MORE
from collections import ChainMap
from inspect import Parameter, Signature
from itertools import chain
from pathlib import Path
from string import punctuation
from typing import Dict, List, Tuple, Optional, Union

from bourbaki.introspection.docstrings import CallableDocs
from ..logging import ProgressLogger, TimedTaskContext
from ..logging.defaults import PROGRESS, ERROR
from ..typed_io.main import ArgSource
from ..typed_io.cli_nargs_ import cli_nargs

VARIADIC_NARGS = (ONE_OR_MORE, OPTIONAL, ZERO_OR_MORE)


class LookupOrderConfigError(ValueError):
    def __init__(self, name_or_names):
        self.args = ("all names in lookup_order must be in {}; got {}".format(tuple(s.name for s in ArgSource), name_or_names),)


class LookupOrderRepeated(LookupOrderConfigError):
    def __init__(self, name_or_names):
        self.args = ("names in lookup_order must be unique; got {}".format(name_or_names),)


_sentinel = object()


class NamedChainMap(ChainMap):
    def __init__(self, *named_maps):
        self.names, maps = zip(*named_maps)
        super().__init__(*maps)

    def get_with_name(self, key, default=_sentinel):
        for name, map_ in zip(self.names, self.maps):
            try:
                return name, map_[key]
            except KeyError:
                continue
        if default is _sentinel:
            raise KeyError(repr(key))

        return None, default


class ReturnHandlerSignatureError(TypeError):
    pass


def strip_command_prefix(prefix: Union[str, Tuple[str, ...]], funcname: str) -> str:
    if isinstance(prefix, str):
        prefix = (prefix,)

    suffix = tuple(p.replace('-', '_') for p in prefix)
    while not funcname.startswith('_'.join(suffix)):
        suffix = suffix[1:]

    strip_prefix = '_'.join(suffix)
    funcname = funcname[len(strip_prefix):].lstrip("_")

    return funcname


def sibling_files(path):
    # for specifying the helper_files arg to the CLI: helper_files=sibling_files(__file__)
    return list(Path(path).parent.glob("*.py"))


def get_task(logger, name, log_level=PROGRESS, error_level=ERROR, time_units='s'):
    if isinstance(logger, ProgressLogger):
        return logger.task(name, time_units=time_units, level=log_level, error_level=error_level)
    return TimedTaskContext(name, logger_or_print_func=logger, time_units=time_units,
                            level=log_level, error_level=error_level)


def update_in(conf, subsection, subconf):
    if not isinstance(subsection, (list, tuple)):
        subsection = (subsection,)

    subconf_ = conf
    for key in subsection:
        subconf__ = subconf_.get(key)
        if subconf__ is None:
            subconf__ = {}
            subconf_[key] = subconf__
        subconf_ = subconf__
    subconf_.update(subconf)


def _help_kwargs_from_docs(docs: CallableDocs, long_desc_as_epilog: bool=False, help_: bool=True):
    if help_:
        kw = {"help": docs.short_desc}
    else:
        kw = {}

    if long_desc_as_epilog:
        kw["description"] = docs.short_desc
        kw["epilog"] = docs.long_desc
    else:
        short = docs.short_desc
        if short:
            if short[-1] in punctuation:
                join = " "
            else:
                join = "; "
            description = join.join(d for d in (short, docs.long_desc) if d)
        else:
            description = docs.long_desc
        kw["description"] = description

    return kw


def _validate_lookup_order(*lookup_order: ArgSource):
    order = []
    missing = []
    for s in lookup_order:
        if isinstance(s, str):
            source = getattr(ArgSource, s, None)
        else:
            source = s

        if not isinstance(source, ArgSource):
            missing.append(s)
        else:
            order.append(source)

    if missing:
        raise LookupOrderConfigError(lookup_order)
    if len(set(lookup_order)) != len(lookup_order):
        raise LookupOrderRepeated(lookup_order)

    return order


def _validate_parse_order(*parse_order):
    if not parse_order:
        raise ValueError("can't make sense of an empty parse_order: {}".format(parse_order))
    if parse_order.count(Ellipsis) > 1:
        raise ValueError("at most one `...` may be present in parse_order; got {}".format(parse_order))
    if not all((isinstance(n, str) or n is Ellipsis) for n in parse_order):
        raise TypeError("all elements of parse_order must be str or `...`; got {}".format(parse_order))
    return parse_order


def _to_name_set(maybe_names, default_set=None, metavar=None):
    if isinstance(maybe_names, str):
        names = {maybe_names}
    elif maybe_names is True:
        # no need to check
        return default_set
    elif not maybe_names:  # None, False
        names = set()
    else:  # collection
        names = set(maybe_names)

    if default_set is not None:
        extra = names.difference(default_set)
        if extra:
            raise NameError("{} must all be in {}; values {} are not"
                            .format(metavar or 'names', default_set, extra))
    return names


def _maybe_bool(names, fallback=_to_name_set):
    if len(names) == 1 and isinstance(names[0], bool):
        return names[0]
    return fallback(names)


def _to_output_sig(output_signature: Signature, function_signature: Optional[Signature] = None,
                   require_keyword_args: bool=False):
    if function_signature is None:
        return output_signature

    common_names = set(output_signature.parameters).intersection(function_signature.parameters)
    if common_names:
        raise ReturnHandlerSignatureError("output handler signature and function signature share arg names: {}"
                                          .format(tuple(common_names)))

    def update_annotation(ann, kind):
        if kind is Parameter.VAR_POSITIONAL:
            return Tuple[ann, ...]
        elif kind is Parameter.VAR_KEYWORD:
            return Dict[str, ann]
        return ann

    if require_keyword_args:
        new_params = [p.replace(kind=Parameter.KEYWORD_ONLY, annotation=update_annotation(p.annotation, p.kind))
                      for p in output_signature.parameters.values()]
        output_signature = Signature(new_params)

    kinds = (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD, Parameter.VAR_POSITIONAL,
             Parameter.KEYWORD_ONLY, Parameter.VAR_KEYWORD)
    params_ = [[], [], [], [], []]
    pos_only, pos, args, kw, kwargs = params_

    params = dict(zip(kinds, params_))
    for sig in (function_signature, output_signature):
        for param in sig.parameters.values():
            params[param.kind].append(param)

    if len(args) > 1:
        extra_args = args.pop()
        extra_args = extra_args.replace(kind=Parameter.KEYWORD_ONLY, annotation=Tuple[extra_args.annotation, ...])
        kw.append(extra_args)

    if len(kwargs) > 1:
        extra_kwargs = kwargs.pop()
        extra_kwargs = extra_kwargs.replace(kind=Parameter.KEYWORD_ONLY, annotation=Dict[str, extra_kwargs.annotation])
        kw.append(extra_kwargs)

    for i in (0, 1):
        params_[i] = _sort_variadic_positionals_last(params_[i])
    all_params = list(chain.from_iterable(params_))

    return Signature(all_params, return_annotation=function_signature.return_annotation, __validate_parameters__=False)


def _sort_variadic_positionals_last(params: List[Parameter]) -> List[Parameter]:
    head, tail = [], []

    for param in params:
        if param.annotation is Parameter.empty and param.default is Parameter.empty:
            head.append(param)
        elif cli_nargs(param.annotation) in VARIADIC_NARGS:
            tail.append(param)
        else:
            head.append(param)

    return head + tail



# coding:utf-8
import os
import ast
import ujson as json
from bourbaki.introspection.prettyprint import fmt_pyobj
from .exceptions import ConfigNotSerializable

DEFAULT_PY_INDENT = '    '
MAX_PY_WIDTH = 80


class UnsafePythonSourceConfig(SyntaxError):
    pass


class IllegalFunctionInSourceConfig(UnsafePythonSourceConfig):
    pass


class IllegalExpressionInSourceConfig(UnsafePythonSourceConfig):
    pass


class PythonConfigNotSerializable(ConfigNotSerializable):
    pass


def is_json_serializable(conf):
    try:
        with open(os.devnull, "w") as f:
            json.dump(conf, f)
    except TypeError:
        return False
    else:
        return True


# safe python syntactic constructs for configuration
legal_python_config_exprs = (ast.Num, ast.Str, ast.BinOp, ast.operator, ast.Compare, ast.Name, ast.Attribute,
                             ast.List, ast.Tuple, ast.Dict, ast.Set, ast.Bytes, ast.Subscript, ast.Ellipsis,
                             ast.ListComp, ast.DictComp, ast.SetComp, ast.comprehension,
                             ast.keyword, ast.NameConstant, ast.Assign, ast.Call)

# safe python builtin functions for configuration
legal_builtin_functions = {'abs', 'all', 'any', 'ascii', 'bin', 'bool', 'bytearray', 'bytes', 'callable', 'chr',
                           'complex', 'dict', 'dir', 'divmod', 'enumerate', 'filter', 'frozenset', 'hash', 'hex',
                           'id', 'int', 'iter', 'len', 'list', 'map', 'max', 'min', 'next', 'oct', 'ord', 'pow',
                           'range', 'repr', 'reversed', 'round', 'set', 'slice', 'sorted', 'str', 'sum', 'tuple',
                           'vars', 'zip'}


def validate_python_config_source(ast_or_source):
    if isinstance(ast_or_source, str):
        tree = ast.parse(ast_or_source)
    else:
        tree = ast_or_source

    msg = ("when using python source as configuration, supply only a single expression "
          "(generally a builtin collection expression) or a sequence of variable assignments; got {}")
    mode = 'exec'
    if len(tree.body) > 1:
        if not all(isinstance(node, ast.Assign) for node in tree.body):
            raise SyntaxError(msg.format(list(map(type, tree.body))))
    elif len(tree.body) == 1:
        if not isinstance(tree.body[0], (ast.Expr, ast.Assign)):
            raise SyntaxError(msg.format(type(tree.body[0])))
        if isinstance(tree.body[0], ast.Expr):
            mode = 'eval'

    def inner(node):
        if isinstance(node, (list, ast.Module)):
            if not isinstance(node, list):
                node = node.body
            return all(inner(v) for v in node)

        if isinstance(node, ast.Call):
            f = node.func
            msg = "function {} is not allowed in python configuration source; occurred at line {}, column {}"
            if isinstance(f, ast.Name) and f.id not in legal_builtin_functions:
                raise IllegalFunctionInSourceConfig(msg.format(repr(node.func.id), node.lineno, node.col_offset))
            elif not isinstance(f, ast.Name):
                raise IllegalFunctionInSourceConfig(msg.format(repr(node.func), node.lineno, node.col_offset))

        if isinstance(node, ast.Expr):
            return inner(node.value)

        if not isinstance(node, legal_python_config_exprs):
            raise IllegalExpressionInSourceConfig("expression type {} is not allowed in python configuration source; "
                                                  "occurred at line {}, column {}"
                                                  .format(type(node), node.lineno, node.col_offset))

        subexprs = tuple(getattr(node, attr) for attr in node._fields)
        subexprs = (e for e in subexprs if isinstance(e, (ast.AST, list)) and not isinstance(e, (ast.Store, ast.Load)))

        return all(map(inner, subexprs))

    if inner(tree):
        if mode == 'eval':
            tree = ast.Expression(tree.body[0].value)
        return tree, mode


def load_python(file):
    source = file.read()
    path = getattr(file, 'name', '<string>')
    tree, mode = validate_python_config_source(source)
    code = compile(tree, filename=path, mode=mode)
    if mode == 'eval':
        conf = eval(source, {})
    else:
        conf = {}
        exec(code, conf)
        conf.pop('__builtins__', None)
    return conf


def dump_python(conf, file, check_json_compat=False):
    if check_json_compat and not is_json_serializable(conf):
        raise PythonConfigNotSerializable("can only write json-serializable config to python source; got {}"
                                         .format(conf))

    print(fmt_pyobj(conf, top_level=True), file=file)



#coding:utf-8
from typing import Tuple, Dict, Callable, Union
from inspect import Signature, Parameter, signature, getmro
from itertools import chain
from logging import Logger, addLevelName, getLoggerClass
from typing import Optional
from . import defaults
from .config import LogLevel, instance_naming_opts
from .helpers import validate_log_level
from .loggers import printLogger, ProgressLogger

addLevelName(defaults.TRACE, defaults.TRACE_LEVEL)


class WithLoggerMixin:
    __logger__ = None
    __logger_cls__ = None
    __log_name__ = None
    __log_level__ = defaults.DEFAULT_CONSOLE_LOG_LEVEL

    @property
    def logger(self):
        logger = self.__dict__.get('__logger__', None)
        if logger is None:
            # construct an instance logger
            instance_name = self.__log_name__
            logger_cls = self.__logger_cls__ or getLoggerClass()
            logger = logger_cls.manager.getLogger(instance_name)
            if logger is not type(self).logger and self.__log_level__ is not None:
                # set an instance level if different than class level
                logger.setLevel(self.__log_level__)
            self.__logger__ = logger
        return logger

    @logger.setter
    def logger(obj, logger):
        assert isinstance(logger, Logger), ("class logger must be an instance of logging.Logger, not {}"
                                            .format(type(logger)))

        obj.__logger__ = logger


# this would ideally bring the metaclass LoggedMeta along for the ride, but in case anyone wants to use it
# to implement the same interface without the potential hiccups of metaclasses, here it is.
# This has the nice side-effect of requiring users of LoggedMeta and its machinery to be explicit that they're using
# a metaclass
class Logged(WithLoggerMixin):
    __verbose__ = defaults.VERBOSE_INSTANCE_INITS
    __instance_naming__ = defaults.DEFAULT_INSTANCE_NAMING
    __name_keyword__ = defaults.DEFAULT_LOG_NAME_KEYWORD
    __last_instance_id__ = None

    def __getstate__(self):
        self.logger.debug("Getting state for pickling/copying")
        # can't pickle RThreadLock objects!
        state = self.__dict__.copy()
        state.pop("__logger__", None)
        self.logger.debug("Got state dict with keys {}".format(sorted(state.keys())))
        return state


missing = object()


class LoggedMeta(type, WithLoggerMixin):
    """
    Metaclass for creating logged classes
    logging-related args can be passed to the metaclass at the top of a class declaration

    example:

    class MyClass(metaclass=LoggedMeta, instance_naming="int"):
        def __init__(self, *args):
            # at this point, the class __new__ has been called and the instance 'self' has a logger
            self.logger.debug("initializing a {} with args {}".format(__class__, args))

    my_instance = MyClass(1, 2, 3)
    my_instance.logger.info("hello!")
    """

    def __new__(mcs, name, bases, namespace,
                level: Optional[LogLevel]=missing,
                use_full_path: bool=False,
                instance_naming: Optional[Union[str, Callable[[object], str]]]=missing,
                name_keyword: str=missing,
                logger_cls: Optional[type]=ProgressLogger,
                verbose: bool=missing):
        """
        Metaclass that gives types/classes and each of their instances a logger (potentially with an
            instance-specific name, depending on the value of instance_naming).

        :param level: the logging level as passed to logging.Logger. messages below this level will be ignored
        :param use_full_path: bool indicating whether to use the full classpath for naming, or just the
            class name. Full class paths can be long and can clutter log files but may help when classes
            have ambiguous names
        :param instance_naming: str or None or callable, one of the following:
            "int": simply use a unique integer to identify new instances, starting from 1
            "hex": use the hexadecimal of the id() function, as used by standard __repr__ method
            "keyword": use a keyword arg from the call to the class constructor
                (this is used implicitly when a value is passed for the name_keyword arg)
            None: don't log from instances separately; just use the class-level logger.
            callable: apply the callable to the named args passed to the constructor to yield a name.
        :param name_keyword: a keyword to use to set the names of instances in the class constructor. This is only used
            when instance_naming is "keyword". This value is appended to the class logger's name following a dash.
            Default is 'log_name'.
        """
        # add Logged as a superclass to allow access to the logger property, allowing lazy logger instantiation
        if Logged not in chain.from_iterable(map(getmro, bases)):
            bases = (Logged, *bases)

        for val, attr in [
            (level, '__log_level__'),
            (instance_naming, '__instance_naming__'),
            (name_keyword, '__name_keyword__'),
            (verbose, '__verbose__'),
            (logger_cls, '__logger_cls__')
        ]:
            # prevents superclass attribute changes from affecting the class - flatten all attr refs from the
            # superclass hierarchy above
            namespace[attr] = namespace_bases_getattr(namespace, bases, attr) if val is missing else val

        validate_log_level(namespace['__log_level__'])
        instance_naming = namespace['__instance_naming__']
        assert instance_naming in instance_naming_opts or callable(instance_naming), \
               "instance_naming must be callable or one of {}; got {}".format(instance_naming_opts, instance_naming)

        name_keyword = namespace['__name_keyword__']
        extra_kw = {name_keyword: str} if instance_naming == 'keyword' else None
        logname = name if not use_full_path else '.'.join((namespace['__module__'], name))
        sig = init_sig(namespace, bases, extra_kw=extra_kw)

        namespace.update(
            __log_name__=logname,
            __last_instance_id__=None,
            # this allows informative introspection in interactive environments such as IPython
            __signature__=sig,
        )

        # construct the class and return it
        cls = type.__new__(mcs, name, bases, namespace)
        return cls

    def __call__(cls: Logged, *args, **kwargs):
        # control instance creation - this is called when you construct an instance of a class
        # this usually happens after logging has been configured, so messages will be faithfully logged
        init_success_msg = "Initialized new instance successfully"
        init_failure_msg = "Instance initialization failed"

        # property lookup on the metaclass, possibly instantiating the logger for the first time
        cls_logger = cls.logger

        if cls.__verbose__:
            cls_logger.debug("constructor called with args ({})".format(call_repr(*args, **kwargs)))
        # create the instance; at this point the class has a logger, which can be used in the cls.__new__ body if needed
        if cls.__new__ is object.__new__:
            obj = cls.__new__(cls)
            new_called = False
        else:
            obj = cls.__new__(cls, *args, **kwargs)
            new_called = True
        cls_logger.debug("Created new instance successfully")
        if new_called:
            cls_logger.debug(init_success_msg)

        naming_convention = cls.__instance_naming__

        if naming_convention is None:
            # then we just use the class logger's name
            instance_name = cls_logger.name
        else:
            if naming_convention == "keyword":
                instance_name = kwargs.get(cls.__name_keyword__)

                if instance_name is None:
                    instance_name = cls_logger.name
                    cls_logger.warning("No name passed to constructor keyword {}; using class logger name"
                                       .format(cls.__name_keyword__))
                else:
                    instance_name = cls_logger.name + str(instance_name)
            elif naming_convention == "int":
                # get and increment the last instance id from the class
                last_id = cls.__last_instance_id__ or 0
                new_id = last_id + 1
                cls.__last_instance_id__ = new_id
                # and name a new instance as '<class log name>-<int id>'
                instance_name = cls_logger.name + "-" + str(new_id)
            elif naming_convention == "hex":
                # use the builtin hex id as implemented by the default __repr__ method
                instance_name = cls_logger.name + "-" + hex(id(obj))
            elif callable(naming_convention):
                sig = signature(cls.__init__)
                params = sig.bind(obj, *args, **kwargs).arguments
                naming_sig = signature(naming_convention)
                new_kw = {k: a for k, a in params.items() if k in naming_sig.parameters}
                instance_name = naming_convention(**new_kw)
            else:
                err_msg = "a logging class's __instance_naming__ attribute must be in {} or callable; got {}" \
                    .format(instance_naming_opts, naming_convention)
                cls_logger.error(err_msg)
                raise ValueError(err_msg)

        setattr(obj, '__log_name__', instance_name)

        init_success_msg += " with log name '{}'".format(instance_name)
        init_failure_msg += " for instance with log name '{}'".format(instance_name)

        # now the instance can construct a logger; it can be used in the instance initialization
        if not new_called:
            try:
                if cls.__init__ is object.__init__:
                    cls.__init__(obj)
                else:
                    cls.__init__(obj, *args, **kwargs)
            except Exception as e:
                cls_logger.error(init_failure_msg, exc_info=True)
                raise e
            else:
                cls_logger.debug(init_success_msg)

        return obj


def short_repr(obj):
    s = repr(obj)
    if len(s) <= defaults.MAX_REPR_LEN:
        return s
    else:
        return s[:defaults.MAX_REPR_LEN // 2] + ' ... ' + s[len(s) - (defaults.MAX_REPR_LEN // 2):len(s)]


def call_repr(*args, **kwargs):
    # print the call signature of a function as it was entered
    return ', '.join(chain(map(short_repr, args),
                           map(lambda tup: tup[0] + "=" + short_repr(tup[1]), kwargs.items())))


def bases_getattr(bases, attr, default=None):
    if not bases:
        return default
    return getattr(bases[0], attr, bases_getattr(bases[1:], attr, default))


def namespace_bases_getattr(namespace, bases, attr, default=None):
    return namespace.get(attr, bases_getattr(bases, attr, default))


def init_sig(namespace: dict, bases: Tuple[type], extra_kw: Optional[Union[Tuple[str], Dict[str, type]]]=None):
    init = namespace_bases_getattr(namespace, bases, '__init__')
    sig = signature(init)
    oldparams = list(sig.parameters.values())[1:]  # throw away 'self' arg for the class signature

    if oldparams and oldparams[-1].kind is Parameter.VAR_KEYWORD:
        varkw = oldparams[-1:]
        oldparams = oldparams[:-1]
    else:
        varkw = []

    if not extra_kw:
        newparams = []
    elif isinstance(extra_kw, tuple):
        newparams = [Parameter(k, kind=Parameter.KEYWORD_ONLY) for k in extra_kw]
    else:
        newparams = [Parameter(k, kind=Parameter.KEYWORD_ONLY, annotation=t) for k, t in extra_kw.items()]

    newsig = Signature(oldparams + newparams + varkw, return_annotation=sig.return_annotation)
    return newsig



#coding:utf-8
import re
from ..reutils import (pathname_re, filename_re, any_name_re, upper_name_re, py_name_dynamic_re,
                              py_dot_name_re, int_re, float_re)


stacktrace_start_re = re.compile(r'^\s*Traceback \(most recent call last\):')

# logging module standard field name -> (printf format type, regex, postprocessor)
log_fields = dict(
    asctime='s',
    created='f',
    filename='s',
    funcName='s',
    levelname='s',
    levelno='d',
    lineno='d',
    module='s',
    msecs='d',
    message='s',
    name='s',
    pathname='s',
    process='d',
    processName='s',
    relativeCreated='d',
    thread='d',
    threadName='s'
)

log_fmt_field_pat = r'(?<!\\)%\(({})\)({})'.format('|'.join(log_fields), '|'.join(set(log_fields.values())))

string_field_regexes = dict(
    filename=filename_re,
    funcName=py_name_dynamic_re,
    levelname=upper_name_re,
    module=py_dot_name_re,
    message=r".*",
    name=any_name_re,
    pathname=pathname_re,
    processName=any_name_re,
    threadName=any_name_re
)

field_regexes = dict(
    f=float_re,
    d=int_re
)

field_converters = dict(
    s=str,
    d=int,
    f=float
)



#coding:utf-8
from logging import NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL
# Useful default configurations

VERBOSE_INSTANCE_INITS = False
VERBOSE_FUNCTION_CALLS = False  # with these true, all logged class inits and logged function
                                # calls are logged with a full signature including the reprs
                                # of the args - good for debug but potentially expensive

MAX_REPR_LEN = 24  # the max string len used for representing args in verbose function call
                   # signature logging

DEFAULT_PROGRESS_LOG_LEVEL = "INFO"

FULL_PATH_LOGGER_NAMING = False  # use the fully qualified class path in the names of class/function
                                 # loggers? Module name is already included in the default log format,
                                 # which makes it an automatically accessible field in a log file parse

DEFAULT_INSTANCE_NAMING = None  # Don't append an instance name to the class logger for new class
                                # instances. Other options are "int", "hex", "keyword"

DEFAULT_LOG_NAME_KEYWORD = "log_name"  # the kwarg that is added to the __init__ signature for
                                       # a class when using the metaclass LoggedMeta

DEFAULT_LOG_DATE_FMT = '%y-%m-%d %H:%M:%S'

VERBOSE_LOG_MSG_FMT = '%(name)s:%(processName)s - %(filename)s:%(funcName)s:%(lineno)d - %(asctime)s,%(msecs)d - %(levelname)s:%(levelno)d - %(message)s'
DEFAULT_LOG_MSG_FMT = '%(name)s - %(asctime)s,%(msecs)d - %(levelname)s:%(levelno)d - %(message)s'
DEFAULT_LOG_MSG_MP_FMT = '%(name)s:%(processName)s - %(asctime)s,%(msecs)d - %(levelname)s:%(levelno)d - %(message)s'


DEFAULT_ROOT_LOG_LEVEL = "INFO"
DEFAULT_CONSOLE_LOG_LEVEL = "INFO"
DEFAULT_FILE_LOG_LEVEL = "INFO"
DEFAULT_SMTP_LOG_LEVEL = "ERROR"
DEFAULT_SMTP_LOG_FLUSH_LEVEL = "CRITICAL"


METALOG_LEVEL = "METALOG"  # level for CountingLogger to use when writing its own log message counts, set above
METALOG = 60               # always log metalog messages
PROGRESS_LEVEL = "PROGRESS"  # level above info for reporting long-running job progress
PROGRESS = 25

TRACE_LEVEL = "TRACE"  # for super fine-grained logging, e.g. of function signatures and return values
TRACE = 1

DEFAULT_FORMATTER_CONFIG = {
                "format": DEFAULT_LOG_MSG_FMT,
                "datefmt": DEFAULT_LOG_DATE_FMT,
            }

DEFAULT_CONSOLE_HANDLER_CONFIG = {
                "class": "logging.StreamHandler",
                "formatter": "default",
                "level": DEFAULT_CONSOLE_LOG_LEVEL,
            }

DEFAULT_CONSOLE_HANDLER_MP_CONFIG = {
                "class": "bourbaki.application.logging.handlers.MultiProcStreamHandler",
                "formatter": "default",
                "level": DEFAULT_CONSOLE_LOG_LEVEL,
            }

# note that the file handler config needs a filename and optionally a maxBytes int
DEFAULT_FILE_HANDLER_CONFIG = {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "default",
                "level": DEFAULT_FILE_LOG_LEVEL,
                "filename": None,
                "maxBytes": None,
            }

DEFAULT_FILE_HANDLER_MP_CONFIG = {
                "class": "bourbaki.application.logging.handlers.MultiProcRotatingFileHandler",
                "formatter": "default",
                "level": DEFAULT_FILE_LOG_LEVEL,
                "filename": None,
                "maxBytes": None,
            }

# note that the smtp handler config needs a capacity, mailhost, toaddrs
DEFAULT_SMTP_HANDLER_CONFIG = {
                "class": "bourbaki.application.logging.handlers.BufferingSMTPHandler",
                "formatter": "default",
                "level": DEFAULT_SMTP_LOG_LEVEL,
                "mailhost": None,
                "credentials": None,
                "toaddrs": None,
                "capacity": None,
                "secure": None,
                "flushLevel": DEFAULT_SMTP_LOG_FLUSH_LEVEL,
            }

DEFAULT_SMTP_HANDLER_MP_CONFIG = {
                "class": "bourbaki.application.logging.handlers.MultiProcBufferingSMTPHandler",
                "formatter": "default",
                "level": DEFAULT_SMTP_LOG_LEVEL,
                "mailhost": None,
                "credentials": None,
                "toaddrs": None,
                "capacity": None,
                "secure": None,
                "flushLevel": DEFAULT_SMTP_LOG_FLUSH_LEVEL,
            }

empty_config_dict = {
        "version": 1,
        "formatters": {
        },
        "handlers": {
        },
        "root": {
            "level": DEFAULT_CONSOLE_LOG_LEVEL
        },
    }



#coding:utf-8
from logging import getLogger, setLoggerClass

# Import commonly used things into the top-level namespace for quick import
from . import config, defaults, interface, timing, loggers, analysis
from .interface import Logged, LoggedMeta
from .config import (configure_default_logging, configure_debug_logging, configure_custom_logging,
                     enable_console_logging, disable_console_logging)
from .defaults import (DEFAULT_LOG_MSG_FMT, DEFAULT_LOG_DATE_FMT,
                       DEFAULT_CONSOLE_LOG_LEVEL, DEFAULT_FILE_LOG_LEVEL)
from .timing import timed_context, TimedTaskContext
from .analysis import log_file_to_df
from .loggers import CallableLogger, ProgressLogger, SwissArmyLogger, printLogger
from .handlers import (MemoryHandler, SMTPHandler, BufferingSMTPHandler, MultiProcHandler,
                       MultiProcBufferingSMTPHandler, MultiProcStreamHandler, MultiProcRotatingFileHandler)


def get_default_logger(name, level=DEFAULT_FILE_LOG_LEVEL, filename=None,
                       dated_logfiles=False):
    """
    Return a logger with the given name, logging to stderr and optionally
    a specified filename.

    Warning: this configures logging with the default configuration as
    a side effect. If you intend to apply a custom configuration, the
    logger returned may be orphaned and/or disabled; you should use
    `configure_default` or `configure_custom`, followed by `getLogger`,
    in that case.
    :param name: the name for the logger. `__file__` is a standard choice
      for a script logger, and `__name__` works within a Jupyter notebook
      session, but isn't very informative - a custom name is better in
      that case
    :param level: The logging level of the root logger through which all
      log messages pass. Set this to "DEBUG" if you want verbose messages.
      Defaults to "INFO".
    :param filename: optional path to a file to log to. Without this, no
      log messages are persisted to disk. Messages logged to this file
      can be loaded as a pandas Dataframe at a later time with
      `log_file_to_df(filename, datetime_index=True/False)`.
    :return: a logging.Logger
    """
    configure_default_logging(filename=filename, level=level, dated_logfiles=dated_logfiles)
    logger = getLogger(name)
    return logger



# coding:utf-8
from typing import Generic, Iterator, Union, Tuple, Mapping, List, Set, Callable, Optional as Opt
from types import FunctionType
from collections import Counter
import os
import sys
import shlex
import shutil
from pathlib import Path
from itertools import chain, repeat
from warnings import warn, filterwarnings
from collections import OrderedDict, ChainMap
from logging import Logger, DEBUG, _levelToName, getLogger
from functools import lru_cache
from inspect import Signature, Parameter
from argparse import (ArgumentParser, Namespace, RawDescriptionHelpFormatter,
                      ONE_OR_MORE, OPTIONAL, ZERO_OR_MORE, SUPPRESS, _SubParsersAction)
from cytoolz import identity, get_in

from bourbaki.introspection.classes import classpath, most_specific_constructor
from bourbaki.introspection.types import (deconstruct_generic, reconstruct_generic, get_param_dict, get_generic_origin,
                                          is_optional_type)
from bourbaki.introspection.typechecking import isinstance_generic
from bourbaki.introspection.imports import LazyImportsCallable
from bourbaki.introspection.docstrings import parse_docstring, CallableDocs
# callables.signature is an lru_cache'ed inspect.signature
from bourbaki.introspection.callables import (signature, fully_concrete_signature, funcname, is_method,
                                              leading_positionals)
from ..completion.completers import CompleteFiles, install_shell_completion
from ..logging import configure_default_logging, Logged, LoggedMeta, ProgressLogger
from ..logging.helpers import validate_log_level_int
from ..logging.defaults import PROGRESS, ERROR, INFO, DEFAULT_LOG_MSG_FMT
from ..config import load_config, dump_config, ConfigFormat, LEGAL_CONFIG_EXTENSIONS
from ..typed_io.utils import to_cmd_line_name, get_dest_name, missing, ellipsis_, text_path_repr
from ..typed_io import TypedIO, ArgSource, CLI, CONFIG, ENV
from .actions import (InfoAction, PackageVersionAction, InstallShellCompletionAction, SetExecuteFlagAction)
from .helpers import (_to_name_set, _validate_parse_order, _help_kwargs_from_docs, _to_output_sig,
                      get_task, NamedChainMap, strip_command_prefix, update_in, _validate_lookup_order)
from .decorators import cli_attrs, NO_OUTPUT_HANDLER

# only need to parse docs once for any function
parse_docstring = lru_cache(None)(parse_docstring)

LOG_LEVELS = sorted(_levelToName, reverse=True)[1:]
LOG_LEVEL_NAMES = [_levelToName[l] for l in LOG_LEVELS]
VARIABLE_LENGTH_NARGS = (ONE_OR_MORE, OPTIONAL, ZERO_OR_MORE)
SUBCOMMAND_ATTR = 'subcommand'
SUBCOMMAND_PATH_ATTR = 'subcommand_path'
CONFIG_FILE_ATTR = 'config_file'
LOGFILE_ATTR = 'logfile'
VERBOSITY_ATTR = 'verbosity'
QUIET_ATTR = 'quiet'
RESERVED_NAMESPACE_ATTRS = (CONFIG_FILE_ATTR, LOGFILE_ATTR, VERBOSITY_ATTR, QUIET_ATTR,
                            SUBCOMMAND_ATTR, SUBCOMMAND_PATH_ATTR)
OUTPUT_GROUP_NAME = 'output control'
OPTIONAL_ARG_TEMPLATE = '{}??'
MIN_VERBOSITY = 1
NoneType = type(None)
ALLOWED_SUBCOMMAND_TYPES = (FunctionType, LazyImportsCallable)
PICKLE_CACHE_SUFFIX = "-cached.pkl"
DEFAULT_EXECUTION_FLAG = '-x'
INSTALL_SHELL_COMPLETION_FLAG = "--install-bash-completion"
CLEAR_CACHE_FLAG = "--clear-cache"
CONFIG_OPTION = "--config"
VERBOSE_FLAGS = ("-v", "--verbose")
QUIET_FLAGS = ("--quiet", "-q")
EXECUTE = False

_type = tuple({type, *(type(t) for t in (Mapping, Tuple, Generic))})


class _DEFAULTS:
    """Stand-in for an ArgSource instance. This is kept out of that enum intentionally since it is always implied;
    we don't want the user to be required to include it when overriding lookup_order since forgetting to do so
    could yield confusing results"""
    value = "function defaults"

    def __str__(self):
        return "{}.{}".format(ArgSource.__name__, "DEFAULTS")


DEFAULTS = _DEFAULTS()

# we call this a lot but not on very many different functions
signature = lru_cache(None)(signature)


# exceptions

class ReservedNameError(AttributeError):
    def __init__(self, reserved, what_names, which_lookup):
        self.args = ("attributes {} are reserved in the parsed argument namespace; {} has attributes {}"
                     .format(reserved, which_lookup, what_names),)


class AmbiguousSignature(TypeError):
    pass


class CLIDefinitionWarning(UserWarning):
    pass


def setup_warn(msg):
    warn(msg, category=CLIDefinitionWarning)


# custom formatters

class MultiplePositionalSequenceArgs(AmbiguousSignature):
    def __init__(self, func_name, names, container_types):
        msg = ("At most one positional arg can have a collection type; otherwise parsing from command line "
               "is ambiguous. Got container types {} for arguments {} of function {}"
               .format(tuple(container_types), tuple(names), func_name))
        self.args = (msg,)


class WideHelpFormatter(RawDescriptionHelpFormatter):
    def __init__(self, prog, indent_increment=2, max_help_position=40, width=None):
        if width is None:
            width = shutil.get_terminal_size().columns
        super().__init__(prog, indent_increment=indent_increment, max_help_position=max_help_position, width=width)


# The main CLI class and helpers


class _SubparserPathAction(_SubParsersAction):
    cmd_prefix = ()

    def add_parser(self, name: str, **kwargs) -> 'PicklableArgumentParser':
        parser = super().add_parser(name, **kwargs)
        parser.cmd_prefix = (*self.cmd_prefix, name)
        return parser

    def __call__(self, parser, namespace, values, option_string, **kwargs):
        cmd = values[0]
        path = (*self.cmd_prefix, cmd)
        setattr(namespace, SUBCOMMAND_PATH_ATTR, path)
        super(_SubparserPathAction, self).__call__(parser, namespace, values, option_string, **kwargs)


class PicklableArgumentParser(ArgumentParser):
    cmd_prefix = ()

    def __init__(self, *args, **kwargs):
        kwargs['formatter_class'] = WideHelpFormatter
        super().__init__(*args, **kwargs)
        # ArgumentParser won't pickle, we have to override its registry here
        self.register('type', None, identity)
        self.register('action', 'parsers', _SubparserPathAction)

    def add_subparsers(self, *args, **kwargs) -> _SubparserPathAction:
        subparsers = super().add_subparsers(*args, **kwargs)
        subparsers.cmd_prefix = self.cmd_prefix
        return subparsers

    @property
    def has_subparsers(self):
        return self._subparsers is not None

    @property
    def subparsers(self):
        if self.has_subparsers:
            return self._subparsers_action
        subparsers = self.add_subparsers(dest=SUBCOMMAND_ATTR,
                                         parser_class=PicklableArgumentParser)
        self._subparsers_action = subparsers
        return subparsers

    def get_nested_subparser(self, *cmd_path: str):
        if not cmd_path:
            return self

        cmdname = cmd_path[0]
        subparsers = self.subparsers
        if subparsers.choices and cmdname in subparsers.choices:
            subparser = subparsers.choices[cmdname]
        else:
            subparser = subparsers.add_parser(cmdname)

        return subparser.get_nested_subparser(*cmd_path[1:])


class CommandLineInterface(PicklableArgumentParser, Logged, metaclass=LoggedMeta):
    """
    Subclass of argparse.ArgumentParser which infers command line interfaces and documentation from functions and
    classes. Type annotations determine parsers for command line args and configuration values, and
    Also adds some functionality to ease use of logging, configuration, verbosity, and dry-run execution.
    """

    lookup_order = ()
    subcommands = None
    init_config_command = None
    app_cls = None
    reserved_attrs = frozenset(RESERVED_NAMESPACE_ATTRS)
    _builtin_commands_added = False
    _unsafe_pickle_attrs = frozenset(("source_file", "helper_files", "default_logfile", "default_configfile",
                                      "_pickle_load_path", "_pickle_dump_path", "_last_edit_time", "_source_files"))
    reserved_command_names = None
    _subparsers_action = None
    _main = None
    parsed = None
    _source_files = None
    _last_edit_time = missing
    _pickle_load_path = None
    _pickle_dump_path = None

    def __init__(self, *,  # <- require keyword args
                 # CLI settings
                 require_keyword_args: bool = True,
                 require_subcommand=False,
                 implicit_flags: bool = False,
                 default_metavars: Opt[Mapping[str, str]] = None,
                 long_desc_as_epilog: bool = True,
                 # config settings
                 use_config_file: Union[bool, str] = False,
                 require_config: bool = False,
                 use_subconfig_for_commands: bool = True,
                 parse_config_as_cli: Union[bool, str, Set[str]] = False,
                 # logging
                 use_logfile: Union[bool, str] = False,
                 log_msg_fmt: str = DEFAULT_LOG_MSG_FMT,
                 dated_logfiles: bool = False,
                 logger_cls: type = ProgressLogger,
                 # logging and config path locations
                 default_paths_relative_to_source=False,
                 # I/O
                 arg_lookup_order: Tuple[ArgSource, ...] = (CLI, ENV, CONFIG),
                 typecheck: bool = False,
                 output_handler: Opt[Callable] = None,
                 # special features and flags
                 use_multiprocessing: bool = False,
                 install_bash_completion: bool = False,
                 use_verbose_flag: bool = False,
                 use_quiet_flag: bool = False,
                 use_execution_flag: Union[bool, str, Tuple[str, ...]] = False,
                 add_install_bash_completion_flag: Union[bool, str] = True,
                 add_init_config_command: Union[bool, str, Tuple[str, ...]] = False,
                 suppress_setup_warnings: bool = False,
                 # source files
                 source_file: Opt[str] = None,
                 helper_files: Opt[List[str]] = None,
                 # info actions
                 version: Opt[Union[str, bool]] = None,
                 package: Opt[str] = None,
                 package_info_keys: Opt[Union[str, Tuple[str, ...]]] = None,
                 # argparse.ArgumentParser init args from here; defaults should be fine in most cases
                 prog: Opt[str] = None,
                 usage: Opt[str] = None,
                 description: Opt[str] = None,
                 epilog: Opt[str] = None,
                 parents=(),
                 formatter_class=WideHelpFormatter,
                 conflict_handler: Opt[str] = 'error',
                 add_help: bool = True,
                 allow_abbrev: bool = True):
        """
        :param require_keyword_args: bool. Should all args be required to be passed from the CLI with an --optional arg,
            (True) or should positional args to functions be interpreted as positional args on the command line (False)?
            The default is True, as this tends to be less error prone and allows more flexibility. If you would like
            positional args however, in functions that do not accept *args, be sure to insert a bare '*' in the
            signature before any args which you would like to be --options.
        :param require_subcommand: Should the app require a subcommand to be passed? The default is False, as this is
            the behavior for simple apps that don't define subcommands. If you are using CommandLineInterface.definition
            on a class however, you should pass True.
        :param implicit_flags: Should boolean non-positional arguments always be interpreted as 0-arg command line
            --flags? In this case, an arg with a False default will be True on passing the flag from the CLI, and an arg
            with a True default will be False on passing the flag. Note however that in this latter case, a '--no-' will
            be prepended to the command line flag to improve readability and semantics.
            Note that when `require_keyword_args` is True, there are no positional args on the command line, so all
            bool-typed args are treated as flags when `implicit_flags` is True.
        :param default_metavars: optional mapping of str -> str. If this is passed, metavars for the command line help
            string are defined by first checking this mapping for the names of function args to be represented before
            falling back to the default `application.typed_io.TypedIO.cli_repr` of the type annotation. To control this
            behavior at the individual function level, use the `application.cli.cli_spec.metavars` decorator or the
            `metavars` arg to `CommandLineInterface.main` or `CommandLineInterface.subcommand`.
        :param long_desc_as_epilog: bool. When True (the default), the `description` arg for argument parsers is defined
            from the "short description" of a function or class, i.e. the portion of the docstring that ends before the
            first double line break, while the `epilog` arg is defined from the "long description" (the part of the
            docstring after the first double line break and before any :param ...: sections. When False, the `epilog`
            arg is always excluded, and the `description` is defined from a concatenation of the long and short
            descriptions.

        :param use_config_file: bool or str. If True, this indicates that a config file can be specified at the command
            line. If a str, this is treated as the default configuration file.
            (see `default_paths_relative_to_source` for path resolution semantics)
        :param require_config: bool. If True, a config file is always required. When `use_config_file` is not a str,
            this implies that a --config file will always be a required arg on the command line.
        :param use_subconfig_for_commands: Should each command get its own subsection in the config, named for the
            function it was defined from? Default is True. If False, you may specify custom subsections for commands by
            using the `application.cli.cli_spec.config_subsections` decorator or directly via the `config_subsections` arg
            to the `CommandLineInterface.main` and `CommandLineInterface.subcommand` decorators; otherwise all args for
            all commands will be taken from the top level namespace of the config.
        :param parse_config_as_cli: bool or set of str. If True, all config values will be parsed with the same parsers
            used on the command line. As configuration is usually a more flexible format than the command line, this is
            hardly ever what you want. If you wish to use the command line parsers for config for a specific set of
            named arguments, pass the set of names, and for all functions that use args with those names, configuration
            values will be parsed using their corresponding CLI parsers.
            Alternately, to control this at the function level, use the `application.cli.cli_spec.parse_config_as_cli`
            decorator or the `parse_config_as_cli` arg of `CommandLineInterface.main` and
            `CommandLineInterface.subcommand` when registering individual functions.

        :param use_logfile: bool or str. If True, a --logfile option will be added to the CLI with no default. If a str,
            a --logfile option will be added to the CLI with this as the default.
            (see `default_paths_relative_to_source` for path resolution semantics)
        :param log_msg_fmt: pass-through to `application.logging.config.configure_default_logging`. The format of logging
            output as understood by the standard library `logging` module.
        :param dated_logfiles: pass-through to `application.logging.config.configure_default_logging`. Should the
            ISO-formatted current datetime be appended to the log file name?
        :param logger_cls: The class to use for all logging. The default is `application.logging.ProgressLogger`.

        :param default_paths_relative_to_source: When this is True (default False), paths supplied for `use_logfile`
            and `use_config_file` are treated as being specified relative to `source_file` (when not absolute or
            relative to $HOME). In this case at parse time, if no command line args are passed for these, the path used
            is expanded to an absolute path, relative to `source_file`. The primary goal of this is to increase
            portability between different environments.

        :param arg_lookup_order: tuple of application.cli.ArgSource enums specifying where args to the executed command
            should be looked up, with earlier entries having higher precedence. Allows the namespace of a configuration
            file or os.environ to populate the args. Default is (CLI, ENV, CONFIG). If CLI is not in the tuple it will
            be inserted as the first entry.

        :param typecheck: bool. Should all parsed args be type-checked before the registered command function is called
            with them? Default is False. Specific args can be typchecked selectively by using the
            `application.cli.typecheck` decorator on registered functions, or passing a list to the `typecheck` arg of
            `CommandLineInterface.main` or `CommandLineInterface.subcommand`.
        :param output_handler: optional callable. Should take the return value of the invoked command/function and
            perform (usually) some IO action on it, such as saving it to disk. The return value is passed as the first
            argument and any further args will be supplied as keyword args parsed from the CLI or config.
        :param use_multiprocessing: bool. If your app uses multiprocessing, then logging will be configured to reflect
            that fact, using appropriate process-safe loggers and handlers. This is passed through to
            `application.logging.config.configure_default_logging` via the `multiprocessing` keyword arg.
        :param install_bash_completion: bool. If True, shell completions are installed automatically at the end of
            interface inference in a `CommandLineInterface.definition` decorator call. If you are registering individual
            functions with `CommandLineInterface.main` or `CommandLineInterface.subcommand`, this option has no effect,
            since it is unknown when the CLI definition is complete. In that case, you can manually call
            `CommandLineInterface.install_shell_completion` in your script. Also see `add_install_bash_completion_flag`.
        :param use_verbose_flag: bool. If passed, a flag is added to the command line interface using the option strings
            `application.cli.VERBOSE_FLAGS` which may be repeated to increase verbosity. This affects verbosity by
             decreasing the logging level by 10 for every repetition. At the DEBUG level (usually 4 repetitions),
             the log format also changes to reflect more information, such as source files and line numbers.
        :param use_quiet_flag: bool. If passed, a flag is added to the command line interface using the option strings
            `application.cli.QUIET_FLAGS` with the effect that when the flag is passed, console logging is suppressed.
        :param use_execution_flag: bool or str. When True or a str, a flag is added to the command line interface which
            is interpreted as specifying that file-system-altering actions may be carried out, with the default behavior
            being to skip these actions, possibly with verbose reporting as in a "dry-run" scenario. Your application
            code may look up the status of this flag via `from application import cli; if cli.EXECUTE: ...` to determine
            what action to take. The only effect that is determined by this class is that file logging is suppressed
            when the flag is passed.
        :param add_install_bash_completion_flag: bool or str. If True or a str, a flag is added to the command line
            interface which triggers installation of bash completions when it is passed. If a str, that flag is equal to
            this arg, else the default flag is `application.cli.INSTALL_SHELL_COMPLETION_FLAG`. Note that the
            'bash-completion' package may need to be installed for your OS for some completions to work; see the
            documentation for `application.completion` for more details.
        :param add_init_config_command: bool or str. When True or a str, a command is added to the command line
            interface which wraps `application.CommandLineInterface.init_config`. This command writes an empty
            configuration file (or dir) for your command line interface that can then be manually edited and passed
            to the --config option when that option is available. See `application.CommandLineInterface.init_config` for
            more details.

        :param suppress_setup_warnings: bool. During processing of these args and inference of the interface from
            annotations and docstrings, some warnings may arise. To suppress these, pass True.

        :param source_file: str, the path to the source where the CLI is defined. This can be accessed simply with the
            module-level variable __file__. This is used to determine last edit time for installation of bash completion
            or caching/inflation of the CLI to/from a pickle. E.g. if automatic installation  of shell completion was
            specified and the source was edited, the shell completions will be reinstalled.
            The base name of this file will also be completed when `install_shell_completion` is True or
            `add_install_bash_completion_flag` is True, upon installation of shell completions.
        :param helper_files: list of str. Any source files which the CLI definition source file imports from, and whose
            edit should trigger repetition of edit-sensitive operations. I.e. this arg serves the same purpose as
            source_file but allows tracking edits in other dependencies.

        :param version: optional str or bool. If str, a --version flag is added that triggers the argparse print version
            action. If bool and `package` is passed, a --version flag is added that prints the version as inferred from
            the passed package.
        :param package_info_keys: optional list of str. When `package` is passed, specifies a subset of the package
            metadata keys to display in the terminal when the --info flag is passed.
        :param package: optional str. If passed, an --info flag is added that prints the metadata for the package as
            found by `pkginfo.get_distribution` and parsed by `pkg_resources.Distribution`. The metadata is printed to
            the terminal in YAML format.

        :param prog: see argparse.ArgumentParser.
            Pass this when `install_shell_completion` is True or `add_install_bash_completion_flag` is True, to specify
            the command name that should be completed.
        :param usage: see argparse.ArgumentParser. This should usually not be passed as it will be informatively
            auto-generated from the registered main function signature.
        :param description: see argparse.ArgumentParser. When no explicitly passed, this is inferred from the docstring
            of the decorated class when `CommandLineInterface.definition` is used, or from the docstring of the
            decorated function when `CommandLineInterface.main` is used. In both cases, this is the "short description",
            i.e. all of the documentation that occurs before the first double line break.
        :param epilog: see argparse.ArgumentParser.  Like `description`, but comes from the "long description" when not
            explicitly passed, i.e. all of the documentation that occurs after the first double line break.

        :param parents: see argparse.ArgumentParser
        :param formatter_class: see argparse.ArgumentParser. We use a custom class `WideHelpFormatter` by default, which
            auto-detects the terminal width and formats the help string to fill it.
        :param conflict_handler: see argparse.ArgumentParser. Default is 'error'
        :param add_help: see argparse.ArgumentParser; default True.
        :param allow_abbrev: see argparse.ArgumentParser; default True.
        """

        super().__init__(prog=prog, usage=usage, description=description, epilog=epilog,
                         parents=parents, prefix_chars='-',
                         fromfile_prefix_chars=None, argument_default=SUPPRESS,
                         formatter_class=formatter_class, conflict_handler=conflict_handler,
                         add_help=add_help, allow_abbrev=allow_abbrev)

        if suppress_setup_warnings:
            filterwarnings("ignore", category=CLIDefinitionWarning)

        if install_bash_completion or add_install_bash_completion_flag:
            if not prog:
                raise ValueError("prog (the program name) must be supplied if "
                                 "add_install_bash_completion_flag/install_bash_completion=True")

        if output_handler is not None and not callable(output_handler):
            raise TypeError("output_handler must be callable; got {}".format(type(output_handler)))

        if isinstance(helper_files, str):
            helper_files = (helper_files,)
        elif helper_files is not None:
            helper_files = tuple(map(str, helper_files))

        if not isinstance(logger_cls, _type) or not issubclass(logger_cls, Logger):
            raise TypeError("logger_cls must be a subclass of {}; got {}".format(Logger, logger_cls))

        # make a mutable set to we can remove reserved namespace attributes as needed
        self.reserved_attrs = set(RESERVED_NAMESPACE_ATTRS)

        if use_verbose_flag:
            self._add_argument(*VERBOSE_FLAGS, action='count', dest=VERBOSITY_ATTR, default=MIN_VERBOSITY,
                               help="specify the level of verbosity; repeat the flag to increase")
        else:
            self.reserved_attrs.remove(VERBOSITY_ATTR)

        if use_quiet_flag:
            self._add_argument(*QUIET_FLAGS, action='store_true', dest=QUIET_ATTR, default=False,
                               help='pass this flag to suppress logging to stdout; '
                                    'This does not effect the verbosity level')
        else:
            self.reserved_attrs.remove(QUIET_ATTR)

        if isinstance(use_logfile, str):
            default_logfile = use_logfile
        else:
            default_logfile = None

        if use_logfile:
            # this will never be a required arg; if not found in config and not given a default in this init,
            # it will default to None, which logging will handle by not configuring a file handler
            self._add_argument('--logfile', action='store', type=str, default=None, dest=LOGFILE_ATTR,
                               help='path to a file to write logs to', completer=CompleteFiles("txt", "log"))
        else:
            self.reserved_attrs.remove(LOGFILE_ATTR)

        if isinstance(use_config_file, str):
            default_configfile = use_config_file
        else:
            default_configfile = None

        if use_config_file or require_config:
            _help = "path to a file to read configuration from"
            if require_config and not isinstance(default_configfile, str):
                kw = dict(required=True)
            else:
                kw = dict(default=None)

            if isinstance(default_configfile, str):
                _help = _help + "; default '{}'.".format(default_configfile)

            self._add_argument(CONFIG_OPTION, type=str, dest=CONFIG_FILE_ATTR, help=_help,
                               metavar=text_path_repr,
                               completer=CompleteFiles(*LEGAL_CONFIG_EXTENSIONS),
                               **kw)
        else:
            self.reserved_attrs.remove(CONFIG_FILE_ATTR)

        if arg_lookup_order:
            self.lookup_order = _validate_lookup_order(*arg_lookup_order)

        if use_execution_flag:
            if isinstance(use_execution_flag, str):
                execution_flag = (use_execution_flag,)
            elif isinstance(use_execution_flag, bool):
                execution_flag = (DEFAULT_EXECUTION_FLAG,)
            else:
                # assume tuple of str
                execution_flag = tuple(use_execution_flag)

            self._add_argument(*execution_flag, action=SetExecuteFlagAction)
        else:
            global EXECUTE
            EXECUTE = True

        if package is not None:
            if version is not None and not isinstance(version, bool):
                setup_warn("both package and version were passed explicitly; the given version will be used rather "
                           "than the version parsed from package info")
                self.add_argument('--version', action='version', version=version)
            elif version or version is None:
                if version is None:
                    version = True
                self.add_argument('--version', action=PackageVersionAction, package=package, version=version)
            self.add_argument('--info', action=InfoAction, package=package, version=version,
                              info_keys=package_info_keys)
        elif version is not None:
            if isinstance(version, bool):
                setup_warn("version={} was passed but no package name was passed; no version can be inferred"
                           .format(version))
            else:
                self.add_argument('--version', action='version', version=version)

        if add_install_bash_completion_flag:
            flag = (add_install_bash_completion_flag if isinstance(add_install_bash_completion_flag, str)
                    else INSTALL_SHELL_COMPLETION_FLAG)
            self._add_argument(flag, action=InstallShellCompletionAction)

        self.version = version
        self.package = package

        self.default_paths_relative_to_source = bool(default_paths_relative_to_source)

        self.use_logfile = bool(use_logfile)
        self.dated_logfiles = bool(dated_logfiles)
        self.default_logfile = default_logfile
        self.log_msg_fmt = log_msg_fmt
        self.app_logger_cls = logger_cls

        self.use_config = bool(use_config_file) or bool(require_config) or bool(add_init_config_command)
        self.use_subconfig_for_commands = bool(use_subconfig_for_commands)
        self.require_config = bool(require_config)
        self.default_configfile = default_configfile
        self.parse_config_as_cli = _to_name_set(parse_config_as_cli, default_set=True, metavar='parse_config_as_cli')
        self.typecheck = bool(typecheck)
        self.output_handler = output_handler

        self.use_multiprocessing = bool(use_multiprocessing)

        self.use_execution_flag = bool(use_execution_flag)
        self.require_subcommand = bool(require_subcommand)
        self.require_keyword_args = bool(require_keyword_args)
        self.implicit_flags = bool(implicit_flags)
        self.default_metavars = None if default_metavars is None else dict(default_metavars)
        self.long_desc_as_epilog = bool(long_desc_as_epilog)

        self.source_file = source_file
        self.helper_files = helper_files
        self._bash_completion = bool(install_bash_completion)
        self.use_init_config_command = bool(add_init_config_command)
        self.subcommands = {}
        self.reserved_command_names = set()
        self._builtin_commands_added = False

        if self.use_config and CONFIG not in self.lookup_order:
            setup_warn("use of configuration was specified with at least one of use_config,"
                       "require_config, or add_init_config_command, but {} is not in lookup_order; "
                       "no value can ever be parsed from configuration in that instance"
                       .format(CONFIG))

        if self.use_init_config_command:
            if isinstance(add_init_config_command, str):
                add_init_config_command = (add_init_config_command,)

            if isinstance(add_init_config_command, tuple):
                self.init_config_command = tuple(map(to_cmd_line_name, add_init_config_command))
            else:
                self.init_config_command = tuple(self.init_config.__name__.split('_'))

            self.reserved_command_names.add(self.init_config_command)

    def add_builtin_commands(self):
        if self._builtin_commands_added:
            return

        if self.use_init_config_command:
            self.subcommand(name=self.init_config_command[-1],
                            command_prefix=self.init_config_command[:-1],
                            output_handler=self.dump_config,
                            require_keyword_args=self.require_keyword_args,
                            implicit_flags=self.implicit_flags,
                            ignore_in_config=True,
                            from_method=False,
                            _builtin=True)(self.init_config)

            path, subcommand = self.get_subcommand_func(self.init_config_command)

            commands_args = [a for a in subcommand.parser._actions if a.dest in ("only_commands", "omit_commands")]
            for arg in commands_args:
                if arg.choices is None:
                    arg.choices = set()
                arg.choices.update(map(' '.join, (t[0] for t in self.all_subcommands() if t[0])))
                arg.metavar = "CMD-NAME"
                choices_str = "{{{}}}".format(",".join(map(shlex.quote, sorted(arg.choices)))) if arg.choices else None
                help = arg.help[arg.help.index(':') + 1:] if arg.help and ':' in arg.help else arg.help
                arg.help = ': '.join(s for s in (choices_str, help) if s)

        self._builtin_commands_added = True
        return self

    def add_argument(self, *args, completer=None, **kwargs):
        dest = kwargs.get('dest', get_dest_name(args, self.prefix_chars))

        if dest in self.reserved_attrs:
            raise KeyError("cannot add arg with options {}; namespace destination '{}' is a reserved namespace "
                           "attribute for this parser".format(args, dest))

        return self._add_argument(*args, completer=completer, **kwargs)

    def _add_argument(self, *args, completer=None, **kwargs):
        action = super().add_argument(*args, **kwargs)
        if completer is not None:
            action.completer = completer

        return action

    def _set_subcommand(self, *cmd_path: str, subcommand: 'SubCommandFunc'):
        cmds = self.subcommands
        for name in cmd_path[:-1]:
            if name not in cmds:
                cmds[name] = None, {}
            _, cmds = cmds[name]

        cmd_name = cmd_path[-1]
        prior_subcmd, prior_subcmds = cmds.get(cmd_name, (None, {}))
        if prior_subcmd is not None and prior_subcmd is not subcommand:
            raise NameError("command {} is already registered to {}".format(cmd_path, prior_subcmd))

        cmds[cmd_name] = subcommand, prior_subcmds

    @property
    def default_prefix_char(self):
        return '-' if '-' in self.prefix_chars else self.prefix_chars[0]

    @property
    def cmd_name(self):
        return self.prog or os.path.split(sys.argv[0])[-1]

    def all_subcommands(self, *prefix: str) -> Iterator[Tuple[Tuple[str, ...], 'SubCommandFunc']]:
        def inner(prefix: Tuple[str, ...], commands: Mapping):
            for name, (cmd, subcmds) in commands.items():
                pre = (*prefix, name)
                if cmd is not None:
                    yield pre, cmd
                yield from inner(pre, subcmds)

        rootcmd = self._main
        subcommands = self.subcommands
        for pre in prefix:
            rootcmd, subcommands = subcommands[pre]

        if rootcmd is not None:
            yield prefix, rootcmd
        yield from inner(prefix, subcommands)

    ##########################
    # main execution methods #
    ##########################

    def run(self, args=None, namespace=None, report_progress=True, time_units='s',
            log_level=PROGRESS, error_level=ERROR):
        ns = self.parse_args(args, namespace)
        cmd, func = self.get_subcommand_func(ns)
        self.logger.debug("command is %r", cmd)
        main = False
        if cmd is None and not self.require_subcommand:
            func = self._main
            if func is None:
                self.error("No subcommand was passed, but no main function has been registered with the "
                           "{}.main() decorator".format(type(self).__name__))
            main = True
            cmd = "MAIN"

        app_logger = self.get_app_logger(ns)
        config = self.parse_config(ns) if self.use_config else None

        app_cls = self.app_cls

        if app_cls is not None and cmd not in self.reserved_command_names:
            # if app_cls.__new__ is not object.__new__ (or lacks the same signature), this should raise
            # an informative error
            app_obj = app_cls.__new__(app_cls)
            if isinstance(app_obj, Logged):
                app_obj.logger = app_logger
            if (not main) and self._main is not None:
                # __init__: perform any initialization logic; if main == True, this will be done below at func.execute()
                self._main.execute(ns, config, app_obj, handle_output=False)
        else:
            app_obj = None

        if not report_progress:
            result = func.execute(ns, config, app_obj)
        else:
            with get_task(app_logger, cmd, log_level=log_level, error_level=error_level, time_units=time_units):
                result = func.execute(ns, config, app_obj)

        return result

    def parse_args(self, args=None, namespace=None):
        if self.require_subcommand and not self.subcommands:
            self.error("This parser requires a subcommand but none have been defined; use the {}.subcommand() "
                       "decorator on a function in your script to define one, or define your CLI via a class"
                       .format(type(self).__name__))

        self.add_builtin_commands()
        ns = super().parse_args(args, namespace=namespace)
        return self.validate_namespace(ns)

    def expand_default_path(self, path):
        path = os.path.expanduser(path)
        if os.path.isabs(path):
            return path
        elif self.default_paths_relative_to_source:
            return os.path.abspath(os.path.join(self.source_file, path))
        else:
            return os.path.abspath(path)

    def validate_namespace(self, ns):
        if self.require_subcommand and self.get_subcommand(ns) is None:
            self.error("A subcommand is required: one of {}".format(tuple(self.subcommands or ())))

        return ns

    @staticmethod
    def get_subcommand(ns=None):
        return getattr(ns, SUBCOMMAND_PATH_ATTR, None)

    def get_subcommand_func(self, ns: Union[Tuple[str, ...], Namespace]) -> Tuple[Tuple[str, ...], 'SubCommandFunc']:
        if isinstance(ns, Namespace):
            cmd_path = self.get_subcommand(ns)
        else:
            cmd_path = ns

        cmd = None
        if cmd_path is not None:
            cmd_path = tuple(cmd_path)
            subcmds = self.subcommands
            for name in cmd_path:
                cmd, subcmds = subcmds[name]

        return cmd_path, cmd

    def get_app_logger(self, ns):
        verbosity = getattr(ns, VERBOSITY_ATTR, 0)
        quiet = getattr(ns, QUIET_ATTR, False)
        log_level_ix = max(MIN_VERBOSITY, min(verbosity, len(LOG_LEVEL_NAMES) - 1))
        log_level = LOG_LEVEL_NAMES[log_level_ix]

        logpath = None
        if self.use_logfile and EXECUTE:
            logpath = getattr(ns, LOGFILE_ATTR, None)
            if logpath is not None:
                logpath = os.path.abspath(logpath)
            elif isinstance(self.default_logfile, str):
                logpath = self.expand_default_path(self.default_logfile)

        self.configure_logging(log_level=log_level, logfile=logpath, quiet=quiet,
                               use_multiprocessing=self.use_multiprocessing)

        rootlogger = getLogger()
        self.logger.debug("configured root logger with effective level %s and handlers %r",
                          _levelToName[rootlogger.getEffectiveLevel()], rootlogger.handlers)

        app_logger = self.app_logger_cls.manager.getLogger(self.cmd_name)
        self.logger.debug("configured app logger with effective level %s and handlers %r",
                          _levelToName[app_logger.getEffectiveLevel()], app_logger.handlers)

        return app_logger

    def configure_logging(self, log_level: Union[int, str], logfile: Opt[str] = None, quiet: bool = False,
                          use_multiprocessing: bool = False):
        log_level_int = validate_log_level_int(log_level)

        configure_default_logging(console=not quiet, filename=logfile,
                                  file_level=log_level, console_level=log_level,
                                  verbose_format=(log_level_int < INFO),
                                  dated_logfiles=self.dated_logfiles,
                                  multiprocessing=use_multiprocessing,
                                  disable_existing_loggers=True)

        self.logger.setLevel(log_level_int)

    def parse_config(self, ns):
        config_file = getattr(ns, CONFIG_FILE_ATTR, None)

        if config_file is None and self.use_config:
            if isinstance(self.default_configfile, str):
                config_file = self.expand_default_path(self.default_configfile)
                if not os.path.exists(config_file):
                    config_file = None
        elif config_file is not None:
            config_file = os.path.abspath(config_file)

        if config_file is None and self.require_config:
            self.error("A config file is required but none was passed and no default was specified")

        no_ext = (not os.path.splitext(config_file)[1]) if isinstance(config_file, (str, Path)) else False

        if config_file is not None:
            # file must exist; this will raise if not
            self.logger.debug("parsing config from {}".format(config_file))
            config = load_config(config_file, namespace=False, disambiguate=no_ext)
        else:
            config = None

        if config is not None:
            self.logger.debug("parsed config:\n%r", config)

        return config

    ##############################
    # command definition methods #
    ##############################

    def subcommand(self, command_prefix=None, config_subsections=None, implicit_flags=None,
                   ignore_on_cmd_line=None, ignore_in_config=None, cmd_line_args=None,
                   parse_config_as_cli=None, parse_order=None, typecheck=None,
                   output_handler=None, named_groups=None, require_keyword_args=None, require_output_keyword_args=None,
                   name=None, from_method=False, metavars=None, tvar_map=None, _main=False, _builtin=False):
        if require_keyword_args is None:
            require_keyword_args = self.require_keyword_args or _main

        if typecheck is None:
            typecheck = self.typecheck

        if implicit_flags is None:
            implicit_flags = self.implicit_flags

        if parse_config_as_cli is None:
            parse_config_as_cli = self.parse_config_as_cli

        if metavars is None:
            metavars = self.default_metavars

        if output_handler is None and not _main:
            output_handler = self.output_handler

        if config_subsections is None:
            if self.use_subconfig_for_commands:
                # bool
                config_subsections = self.use_config
            else:
                config_subsections = [()]

        kw = dict(argparser_cmd_name=self.cmd_name,
                  command_prefix=command_prefix,
                  config_subsections=config_subsections,
                  implicit_flags=implicit_flags,
                  require_keyword_args=require_keyword_args,
                  require_output_keyword_args=require_output_keyword_args,
                  ignore_on_cmd_line=ignore_on_cmd_line,
                  ignore_in_config=ignore_in_config,
                  parse_config_as_cli=parse_config_as_cli,
                  cmd_line_args=cmd_line_args,
                  named_groups=named_groups,
                  typecheck=typecheck,
                  output_handler=output_handler,
                  name=name,
                  from_method=from_method,
                  lookup_order=self.lookup_order,
                  parse_order=parse_order,
                  metavars=metavars,
                  tvar_map=tvar_map,
                  _main=_main)

        def dec(f):
            subcmd = SubCommandFunc(f, **kw)

            if self.reserved_command_names and not _builtin and subcmd.cmd_prefix in self.reserved_command_names:
                raise NameError("cannot use name '{}' for a subcommand{}; it is reserved for builtin functionality"
                                .format(name,
                                        " under prefix {}".format(command_prefix) if command_prefix else ''))

            subcmd.parser = self.add_arguments_from(subcmd)
            return subcmd

        return dec

    def main(self, config_subsections=None, implicit_flags=None,
             ignore_on_cmd_line=None, ignore_in_config=None,
             cmd_line_args=None, parse_config_as_cli=None, parse_order=None, typecheck=None,
             output_handler=None, named_groups=None,
             name=None, from_method=False, metavars=None, tvar_map=None):
        return self.subcommand(config_subsections=config_subsections,
                               implicit_flags=implicit_flags,
                               require_keyword_args=True,
                               ignore_on_cmd_line=ignore_on_cmd_line,
                               ignore_in_config=ignore_in_config,
                               parse_config_as_cli=parse_config_as_cli,
                               cmd_line_args=cmd_line_args,
                               typecheck=typecheck,
                               output_handler=output_handler,
                               named_groups=named_groups,
                               parse_order=parse_order,
                               metavars=metavars,
                               name=name,
                               from_method=from_method,
                               tvar_map=tvar_map, _main=True)

    def definition(self, app_cls: type):
        """class decorator for generating subcommands via a class.
        This should only be called once per instance."""
        if not isinstance(app_cls, _type):
            t = type(self)
            raise TypeError("{} instances can only be used to decorate classes; if you would like to customize "
                            "behavior for specific subcommands, use {}.subcommand or {}.main as decorators for "
                            "functions".format(classpath(t), t.__name__, t.__name__))

        app_cls_ = get_generic_origin(app_cls)
        self.app_cls = app_cls_

        if getattr(app_cls_, "__doc__", None):
            if self.description is None or self.epilog is None:
                docs = parse_docstring(app_cls_.__doc__)
                help_kw = _help_kwargs_from_docs(docs, long_desc_as_epilog=self.long_desc_as_epilog, help_=False)
                for name, value in help_kw.items():
                    if getattr(self, name, None) is None:
                        setattr(self, name, value)

        tvar_map = get_param_dict(app_cls)
        have_default_config_as_cli_args = not isinstance(self.parse_config_as_cli, bool)
        have_default_metavars = isinstance(self.default_metavars, dict)

        # do the init last if it wasn't in the class namespace
        for name, f in chain([("__init__", most_specific_constructor(app_cls))], app_cls_.__dict__.items()):
            if not callable(f):
                continue
            if cli_attrs.noncommand(f):
                continue
            if name.startswith("_") and name != "__init__":
                # don't register implicitly private methods
                continue
            if not is_method(app_cls, name):
                # don't configure CLI commands for classmethods and staticmethods
                setup_warn("currently only basic methods are supported as subcommands in a class def context "
                           "but callable '{}' = {} defined in {}'s namespace is of type {}"
                           .format(name, f, app_cls, type(f)))
                continue
            if f is object.__init__:
                # don't configure CLI for base object constructor
                continue

            # function decorators override CLI configuration
            parse_config_as_cli_ = cli_attrs.parse_config_as_cli(f, self.parse_config_as_cli)
            if have_default_config_as_cli_args and not isinstance(parse_config_as_cli_, bool):
                if parse_config_as_cli_ is not self.parse_config_as_cli:
                    parse_config_as_cli_ = self.parse_config_as_cli.union(parse_config_as_cli_)

            metavars = cli_attrs.metavars(f, self.default_metavars)
            if have_default_metavars and metavars is not self.default_metavars:
                metavars = ChainMap(metavars, self.default_metavars)

            require_output_kwargs = cli_attrs.require_output_keyword_args(f, self.require_keyword_args)

            if not self.use_config:
                default_subsections = False
            else:
                default_subsections = True if self.use_subconfig_for_commands else [()]

            extra_kw = dict(
                implicit_flags=self.implicit_flags,
                from_method=True,
                typecheck=cli_attrs.typecheck(f),
                ignore_on_cmd_line=cli_attrs.ignore_on_cmd_line(f, False),
                ignore_in_config=cli_attrs.ignore_in_config(f),
                config_subsections=cli_attrs.config_subsections(f, default_subsections),
                parse_config_as_cli=parse_config_as_cli_,
                parse_order=cli_attrs.parse_order(f),
                named_groups=cli_attrs.named_groups(f),
                metavars=metavars,
                tvar_map=tvar_map,
            )

            if name == "__init__" and self._main is None:
                # if we hit the init but didn't register one yet
                self.main(**extra_kw)(f)
            elif name != "__init__":
                require_keyword_args = cli_attrs.require_keyword_args(f, self.require_keyword_args)
                self.subcommand(require_keyword_args=require_keyword_args,
                                require_output_keyword_args=require_output_kwargs,
                                output_handler=cli_attrs.output_handler(f, self.output_handler),
                                name=name,
                                command_prefix=cli_attrs.command_prefix(f),
                                **extra_kw)(f)

        if self.description is None:
            self.description = app_cls.__doc__

        self.add_builtin_commands()

        if self._bash_completion:
            self.install_shell_completion()

        return app_cls

    def add_arguments_from(self, subcmd_func: 'SubCommandFunc') -> PicklableArgumentParser:
        cmd_path = (*subcmd_func.cmd_prefix, subcmd_func.cmd_name)

        if not subcmd_func._main:
            subparsers = self.get_nested_subparser(*subcmd_func.cmd_prefix).subparsers
            docs = subcmd_func.docs
            if docs is None:
                parser = subparsers.add_parser(subcmd_func.cmd_name)
            else:
                help_kw = _help_kwargs_from_docs(docs, long_desc_as_epilog=self.long_desc_as_epilog)
                parser = subparsers.add_parser(subcmd_func.cmd_name, **help_kw)

            self._set_subcommand(*cmd_path, subcommand=subcmd_func)
        else:
            parser = self.add_argument_group("main arguments")

            if self.description is None:
                self.description = subcmd_func.docs.short_desc
                if self.epilog is None:
                    self.epilog = subcmd_func.docs.long_desc

            self._main = subcmd_func

        self._add_arguments_from(parser, subcmd_func)
        return parser

    @staticmethod
    def _add_arguments_from(parser, subcmd_func: 'SubCommandFunc'):
        params = subcmd_func.signature.parameters
        allow_positionals = not subcmd_func.require_keyword_args
        arg_to_group = subcmd_func.arg_to_group_name
        named_groups = {name: parser.add_argument_group(name) for name in subcmd_func.named_groups}
        lookup_order = (CLI, *subcmd_func.lookup_order, DEFAULTS) if CLI not in subcmd_func.lookup_order \
            else (*subcmd_func.lookup_order, DEFAULTS)

        positional_args = []
        for name, io_methods in subcmd_func.typed_io.items():
            if name in subcmd_func.ignore_on_cmd_line:
                continue
            param = params[name]
            group_name = arg_to_group.get(name)
            group = parser if group_name is None else named_groups[group_name]

            has_fallback = False
            if CONFIG in lookup_order and (name not in subcmd_func.ignore_in_config):
                has_fallback = True
            if ENV in lookup_order and (name in subcmd_func.parse_env):
                has_fallback = True

            action = io_methods.add_argparse_arg(group, param,
                                                 allow_positionals=allow_positionals,
                                                 implicit_flags=subcmd_func.implicit_flags,
                                                 has_fallback=has_fallback,
                                                 metavar=subcmd_func.metavars,
                                                 docs=subcmd_func.docs)
            nargs = io_methods.cli_nargs
            if action.positional:
                positional_args.append((name, param.annotation, nargs))

        if positional_args:
            # one trailing variable-length positional is OK
            bad_positional_args = [tup for tup in positional_args[:-1] if tup[-1] in VARIABLE_LENGTH_NARGS]
            if bad_positional_args:
                if positional_args[-1][-1] in VARIABLE_LENGTH_NARGS:
                    bad_positional_args.append(positional_args[-1])
                raise ValueError("The parameters {} of function {} are all positional, but have variable-length "
                                 "command line args {}; parsing cannot be performed unambiguously."
                                 .format(", ".join("{}:{}".format(n, t) for n, t, _ in bad_positional_args),
                                         subcmd_func.func_name, tuple(n for _, _, n in bad_positional_args)))


    #######################
    # Source file helpers #
    #######################

    def source_files(self):
        if self._source_files is not None:
            return self._source_files
        sourcepath = self.get_sourcepath()

        helpers = []
        for helper in self.helper_files or ():
            if helper is not None:
                path = os.path.abspath(helper)
                if not os.path.exists(path):
                    setup_warn("helper file {} was specified but doesn't exist".format(helper))
                else:
                    helpers.append(path)

        paths = [sourcepath, *helpers] if sourcepath else helpers
        self._source_files = paths
        return paths

    def get_sourcepath(self):
        if self.source_file is None:
            import __main__ as main
            sourcepath = getattr(main, '__file__', None)
            if sourcepath is not None:
                sourcepath = os.path.abspath(sourcepath)
                if not os.path.exists(sourcepath):
                    setup_warn("source file {} was looked up on __main__ but doesn't exist".format(sourcepath))
        else:
            sourcepath = os.path.abspath(self.source_file)
        return sourcepath

    def last_edit_time(self):
        if self._last_edit_time is not missing:
            return self._last_edit_time

        sources = self.source_files()
        time = None if not sources else max(os.stat(path).st_mtime for path in sources)
        self._last_edit_time = time
        return time

    ####################
    # Shell completion #
    ####################

    def install_shell_completion(self):
        if self.source_file is None:
            names = (self.prog,)
        else:
            names = (self.prog, os.path.basename(self.source_file))

        self.logger.debug("Installing bash completion for command `{}`".format(names))
        install_shell_completion(self, *names, last_edit_time=self.last_edit_time())

    ##################
    # pickle methods #
    ##################

    def __getstate__(self):
        state = super().__getstate__()
        app_cls = state.get("app_cls", None)
        if app_cls is not None:
            state["app_cls"] = deconstruct_generic(app_cls)
        state.pop("_source_files", None)  # could change in a new context if relative paths or ~/ are used
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        app_cls = state.get("app_cls", None)
        if app_cls is not None:
            app_cls = reconstruct_generic(app_cls)
            self.app_cls = app_cls

    ########################
    # special CLI commands #
    ########################

    def empty_config(self, only_required_args=False, literal_defaults=True, only_commands=None, omit_commands=None):
        config = {}
        MAIN = ()
        for name, subcommand in chain([(MAIN, self._main)], self.all_subcommands()):
            if only_commands is not None and name not in only_commands and name is not MAIN:
                continue
            if omit_commands is not None and name in omit_commands:
                continue
            if not subcommand.config_subsections:
                continue

            subsection = subcommand.config_subsections[0]

            if name != MAIN:
                self.logger.info("computing empty configuration for command '{}', subsection {}"
                                 .format(' '.join(name), subsection))
            else:
                self.logger.info("computing empty configuration for main args" +
                                 (", subsection {}".format(subsection) if subsection else " at top level"))

            subconf = subcommand.empty_config(only_required_args=only_required_args,
                                              literal_defaults=literal_defaults)
            self.logger.info('%r', subconf)
            update_in(config, subcommand.config_subsections[0], subconf)

        return config

    def init_config(self, *,
                    only_required_args: bool = False,
                    literal_defaults: bool = True,
                    only_commands: Opt[Set[str]] = None,
                    omit_commands: Opt[Set[str]] = None):
        """
        Initialize a configuration for this app.

        :param only_required_args: should only required arguments be given and empty config value?
        :param literal_defaults: should literal default values be written to the config?
        :param only_commands: only write configuration for these commands
        :param omit_commands: omit configuration for these commands
        """
        # this is added as a CLI command if specified in the constructor
        self.logger.info("constructing empty configuration")
        if only_commands:
            self.logger.info("only producing configuration for commands {}".format(only_commands))
        elif omit_commands:
            self.logger.info("omitting commands {}".format(omit_commands))

        only_commands = [tuple(s.split()) for s in only_commands] if only_commands else None
        omit_commands = [tuple(s.split()) for s in omit_commands] if omit_commands else None

        config = self.empty_config(only_required_args=only_required_args, literal_defaults=literal_defaults,
                                   only_commands=only_commands, omit_commands=omit_commands)
        self.logger.debug("new config has keys {}".format(config.keys()))
        return config

    def dump_config(self, config: Mapping,
                    path: Opt[Path] = None,
                    *,
                    as_dir: bool = False,
                    format: Opt[ConfigFormat] = None,
                    use_default_path: bool = False):
        """
        Optionally save the configuration to a file, or print to stdout if no path is provided.

        :param config: the configuration to dump (usually a JSON-serializable python object)
        :param path: path to the file (or dir) to write the configuration to. '-' implies stdout; note that in this case
            a format must also be specified, since it cannot be inferred from the filename.
        :param as_dir: should the config be written to a dir? (one file per subsection)
        :param format: the file extension to use for the configuration; determines the markup language used.
           Note: the .ini format is constrained to 1-level nesting and cannot represent values needing more than this
        :param use_default_path: use the default config file path for this app.
        """
        if use_default_path:
            if path is not None:
                raise ValueError("Cannot supply both a path and specify default_path=True")
            if os.path.exists(self.default_configfile):
                msg = "Config file already exists at default location {}".format(self.default_configfile)
                warn(msg)
                if input("Overwrite (y/n)? ").strip().lower() == 'y':
                    path = self.default_configfile
                else:
                    exc = FileExistsError(msg)
                    self.logger.error(msg, exc_info=exc)
                    raise exc
            else:
                path = self.default_configfile

        file = sys.stdout if path is None or str(path) == "-" else path

        self.logger.info("writing new configuration to {}{}".format(
            file, "" if format is None else " with {} format".format(format.name)))

        try:
            dump_config(config, file, ext=format, as_dir=as_dir)
        except Exception as e:
            self.logger.exception("could not dump config to {}".format(file))
            raise e


class SubCommandFunc(Logged):
    __log_level__ = DEBUG
    parser = None

    def __init__(self,
                 func: Callable, *,
                 argparser_cmd_name=None,
                 lookup_order=None,
                 named_groups=None,
                 implicit_flags=False,
                 command_prefix=None,
                 config_subsections=None,
                 ignore_on_cmd_line=None,
                 cmd_line_args=None,
                 parse_env=None,
                 ignore_in_config=None,
                 parse_config_as_cli=None,
                 require_keyword_args=True,
                 require_output_keyword_args=True,
                 parse_order=None,
                 typecheck=False,
                 output_handler=None,
                 name=None,
                 from_method=False,
                 metavars=None,
                 tvar_map=None,
                 _main=False):
        if name is None:
            func_name = funcname(func)
            if func_name is None:
                raise AttributeError("functions for subcommands must have __name__ attributes; {} does not; "
                                     "try using a def rather than a lambda expression / funtools.partial, or pass a "
                                     "name arg explicitly".format(func))
        else:
            func_name = name

        if isinstance(command_prefix, str):
            command_prefix = (command_prefix,)
        elif command_prefix is None:
            command_prefix = ()

        if not named_groups:
            named_groups = {}

        lookup_order = _validate_lookup_order(*lookup_order)
        cmd_name = to_cmd_line_name(strip_command_prefix(command_prefix, func_name))
        command_prefix = tuple(map(to_cmd_line_name, command_prefix))

        if ignore_on_cmd_line is not None and cmd_line_args is not None:
            raise ValueError("At most one of `ignore_on_cmd_line`, `cmd_line_args` should be passed")

        try:
            docs = parse_docstring(func)
        except AttributeError:
            docs = None

        cli_sig = fully_concrete_signature(func, from_method=from_method, tvar_map=tvar_map)
        # non-output_handler positionals
        positional_names = leading_positionals(cli_sig.parameters, names_only=True)

        if output_handler is NO_OUTPUT_HANDLER:
            output_handler = None

        if output_handler is not None:
            # from_method = True because we skip the first arg
            output_sig = fully_concrete_signature(output_handler, from_method=True, tvar_map=tvar_map)
            output_param_names = set(output_sig.parameters)
            output_positional_names = leading_positionals(output_sig.parameters, names_only=True)
            # passing require_keyword_args prevents potential superfluous signature errors
            cli_sig = _to_output_sig(output_sig, cli_sig,
                                     require_keyword_args=require_keyword_args or require_output_keyword_args)
            try:
                output_docs = parse_docstring(output_handler)
            except AttributeError:
                output_docs = None
            else:
                docs = CallableDocs(short_desc=docs.short_desc,
                                    long_desc='\n'.join(d for d in (docs.long_desc, output_docs.desc) if d),
                                    params=chain(docs.params.values(), output_docs.params.values()),
                                    returns=docs.returns,
                                    raises=docs.raises + output_docs.raises)

            if OUTPUT_GROUP_NAME in named_groups:
                raise NameError("'{}' is a group name reserved for output handler args, when an output handler "
                                "is specified".format(OUTPUT_GROUP_NAME))
            else:
                named_groups[OUTPUT_GROUP_NAME] = output_param_names
        else:
            output_docs = None
            output_param_names = set()
            output_positional_names = ()

        # all parameter names, including output_handler if present
        param_names = set(cli_sig.parameters.keys())

        defaults = {}
        for name, p in cli_sig.parameters.items():
            if p.default is not Parameter.empty:
                defaults[name] = p.default
            elif p.kind == Parameter.VAR_POSITIONAL:
                defaults[name] = ()
            elif p.kind == Parameter.VAR_KEYWORD:
                defaults[name] = {}

        if isinstance(cmd_line_args, str):
            cmd_line_args = {cmd_line_args}
        elif not cmd_line_args:
            cmd_line_args = set()

        self._main = bool(_main)
        self.func_name = func_name
        self.cmd_prefix = command_prefix
        self.cmd_name = cmd_name
        self.__log_name__ = ".".join((argparser_cmd_name, *self.cmd_prefix, self.cmd_name))
        self.docs = docs
        self.output_docs = output_docs
        self.func = func
        self.output_handler = output_handler
        self.from_method = bool(from_method)

        # parameter names
        self.param_names = param_names
        self.positional_names = positional_names
        self.output_param_names = output_param_names
        self.output_positional_names = output_positional_names
        self.defaults = defaults
        self.signature = cli_sig
        self.implicit_flags = bool(implicit_flags)
        self.require_keyword_args = bool(require_keyword_args)
        self.require_output_keyword_args = bool(require_output_keyword_args)
        self.lookup_order = lookup_order

        # these all require validation against the set of argument names
        # NOTE: order is important here! some things use properties, so that the self.<name> reference later
        # may be a validated version
        self.ignore_on_cmd_line = _to_name_set(
            ignore_on_cmd_line,
            default_set=param_names.difference(cmd_line_args),
            metavar='ignore_on_cmd_line',
        )
        self.ignore_in_config = _to_name_set(
            ignore_in_config,
            default_set=param_names,
            metavar='ignore_in_config',
        )
        self.parse_config_as_cli = _to_name_set(
            parse_config_as_cli,
            default_set=param_names,
            metavar='parse_config_as_cli',
        )
        self.typecheck = _to_name_set(
            typecheck,
            default_set=param_names,
            metavar='typecheck',
        )
        self.parse_env = parse_env
        self.parsed_param_names = {
            name for name in self.param_names
            if name not in self.ignore_in_config
            or name not in self.ignore_on_cmd_line
            or name in self.parse_env
        }
        self.parse_order = self.compute_parse_order(parse_order, self.parsed_param_names)
        self.metavars = metavars
        self.named_groups = named_groups
        self.arg_to_group_name = dict(
            chain.from_iterable(zip(argnames, repeat(groupname))
                                for groupname, argnames in self.named_groups.items())
        )

        self.typed_io = OrderedDict([(name, TypedIO.from_parameter(param))
                                     for name, param in cli_sig.parameters.items()
                                     if name in self.parsed_param_names])

        if config_subsections in (False, None):
            self.config_subsections = None
        elif config_subsections is True:
            # preserve _'s on __init__ for main function
            self.config_subsections = [(*command_prefix, func_name if _main else cmd_name)]
        elif isinstance(config_subsections, (str, int)):
            self.config_subsections = [(config_subsections,)]
        elif isinstance(config_subsections, tuple):
            # a single section
            self.config_subsections = [config_subsections]
        else:
            # a list of sections
            self.config_subsections = [(t,) if isinstance(t, (str, int)) else t for t in config_subsections]

    @property
    def named_groups(self):
        return self._named_groups

    @named_groups.setter
    def named_groups(self, named_groups: Opt[Mapping[str, Set[str]]]):
        if not named_groups:
            named_groups = {}
        else:
            named_groups = {name: _to_name_set(argnames, self.param_names, metavar="args in named group '{}'".format(name))
                            for name, argnames in named_groups.items()}
            counts = Counter(chain.from_iterable(named_groups.values()))
            repeated = [argname for argname, count in counts.items() if count > 1]
            if repeated:
                raise ValueError(
                    "arguments {} were repeated across {} named groups respectively for command with function name '{}'"
                    .format(list(repeated), [counts[n] for n in repeated], self.func_name)
                )
        self._named_groups = named_groups

    @property
    def parse_env(self):
        return self._parse_env

    @parse_env.setter
    def parse_env(self, parse_env: Opt[Mapping[str, str]]):
        self._parse_env = self._validate_arg_keys(parse_env, 'parse_env')

    def _validate_arg_keys(self, dict_, group_name):
        # simply throw an error if there are unknown keys
        _ = _to_name_set(dict_, self.param_names, metavar=group_name)
        return dict_ or {}

    @property
    def cmd_path(self):
        return (*self.cmd_prefix, self.cmd_name)

    @staticmethod
    def compute_parse_order(parse_order, param_names):
        if parse_order is None:
            return list(param_names)
        parse_order = _validate_parse_order(*parse_order)

        if any((n not in param_names) or (n is not Ellipsis) for n in parse_order):
            raise NameError("parse_order entries must all be in {}; got {}".format(param_names, parse_order))

        try:
            ellipsis_ix = parse_order.index(Ellipsis)
        except ValueError:
            head, tail = parse_order, ()
            middle = ()
        else:
            head, tail = parse_order[:ellipsis_ix], parse_order[ellipsis_ix + 1:]
            middle = param_names.difference(head).difference(tail)

        return [*(h for h in head if h in param_names), *middle, *(t for t in tail if t in param_names)]

    def get_conf(self, config):
        if config is None:
            return None

        conf = None
        ignore = self.ignore_in_config or ()

        if self.config_subsections:
            confs = []
            for section in self.config_subsections:
                conf_ = get_in(section, config, missing)

                if conf_ is not missing:
                    if not isinstance(conf_, Mapping):
                        raise TypeError('config subsections for function arguments must be str->value mappings; got {} '
                                        'for command {} in section {} of the config'
                                        .format(type(conf_), self.__name__, section))
                    confs.append(conf_)

            if len(confs) > 1:
                conf = ChainMap(*confs)
            elif not confs:
                conf = {}
            else:
                conf = confs[0]

            if conf and ignore:
                conf = {k: v for k, v in conf.items() if k not in ignore}

        return conf

    def get_env(self):
        if not self.parse_env:
            return None

        env = {}
        for arg_name, env_name in self.parse_env.items():
            if env_name in os.environ:
                env[arg_name] = os.environ[arg_name]
        return env

    def execute(self, namespace, config, instance=None, handle_output=True):
        args, kwargs, output_args, output_kwargs = self.prepare_args_kwargs(namespace, config)

        if self.from_method:
            # method
            value = self.func(instance, *args, **kwargs)
        else:
            # bare function
            value = self.func(*args, **kwargs)

        if handle_output and self.output_handler is not None:
            output_args = output_args or ()
            output_kwargs = output_kwargs or {}
            self.output_handler(value, *output_args, **output_kwargs)

        return value

    def prepare_args_kwargs(self, argparse_namespace, config=None):
        conf = self.get_conf(config)
        env = self.get_env()
        cli = argparse_namespace.__dict__ if isinstance(argparse_namespace, Namespace) else argparse_namespace
        cli = {name: value for name, value in cli.items() if value is not missing}

        lookup = {CLI: cli, ENV: env, CONFIG: conf, DEFAULTS: self.defaults}
        lookup_order = (CLI, *self.lookup_order, DEFAULTS) if CLI not in self.lookup_order else (*self.lookup_order, DEFAULTS)
        sources = []
        for source in lookup_order:
            ns = lookup[source]
            if ns:
                sources.append((source, ns))

        namespace = NamedChainMap(*sources)
        self.logger.debug("parsing args for command %r from sources %r in order %r",
                          self.func_name, namespace.names, self.parse_order)

        values = []
        missing_ = []
        sentinel = object()
        for name in self.parse_order:
            source, value = namespace.get_with_name(name, sentinel)
            if value is sentinel:
                missing_.append(name)
            else:
                values.append((name, source, value))

        if missing_:
            raise ValueError("could not find value for args {} of command `{}` in any of {}"
                             .format(missing_, self.cmd_name, tuple(source.value for source in lookup_order)))

        typed_io = self.typed_io
        params = self.signature.parameters
        handle_output = self.output_handler is not None

        final_args = ()
        final_kwargs = {}
        final_output_args = () if handle_output else None
        final_output_kwargs = {} if handle_output else None
        for name, source, value in values:
            self.logger.debug("parsing arg %r from %s with value %r", name, source.value, value)
            tio = typed_io[name]
            is_output_arg = name in self.output_param_names

            if handle_output and is_output_arg:
                args, kwargs = final_output_args, final_output_kwargs
            else:
                args, kwargs = final_args, final_kwargs

            if source == CONFIG and name in self.parse_config_as_cli:
                parser = tio.cli_parser
            else:
                parser = tio.parser_for_source(source)

            param = params[name]
            kind = param.kind
            parsed = parser(value)
            self.logger.debug("parsed value %r for arg %r from %s", parsed, name, source.value)

            if name in self.typecheck and param.annotation is not Parameter.empty:
                if not isinstance_generic(parsed, tio.type_):
                    try:
                        raise TypeError("parsed value {} for arg {} is not an instance of {}"
                                        .format(repr(parsed), repr(name), param.annotation))
                    except Exception as e:
                        self.logger.exception()
                        raise e

            if kind == Parameter.VAR_POSITIONAL:
                # only use *args for the args to start, then extend named positionals with these below
                if is_output_arg:
                    final_output_args = parsed
                else:
                    final_args = parsed
            elif kind == Parameter.VAR_KEYWORD:
                kwargs.update(parsed)
            else:
                kwargs[name] = parsed

        if self.positional_names:
            final_args = (*(final_kwargs.pop(n) for n in self.positional_names), *final_args)
        if self.output_positional_names:
            final_output_args = (
                *(final_output_kwargs.pop(n) for n in self.output_positional_names),
                *final_output_args
            )

        return final_args, final_kwargs, final_output_args, final_output_kwargs

    def empty_config(self, only_required_args=False, literal_defaults=False):
        conf = {}
        typed_io = self.typed_io
        for name, param in self.signature.parameters.items():
            if name in self.ignore_in_config:
                continue

            has_default = param.default is not Parameter.empty
            if has_default and only_required_args:
                continue

            # for the purposes of a representative config value, None doesn't count as a default
            has_default = has_default and param.default is not None

            tio = typed_io[name]
            if name in self.parse_config_as_cli:
                # can't cli-encode a value, so even if literal_defaults is True, we leave this as a type repr
                val = tio.cli_repr
                if tio.cli_nargs in (ZERO_OR_MORE, ONE_OR_MORE):
                    if isinstance(val, str):
                        val = [val, ellipsis_]
                    elif not isinstance(val, list):
                        val = list(val)
            elif literal_defaults and has_default:
                val = tio.config_encoder(param.default)
            else:
                val = tio.config_repr

            if is_optional_type(tio.type_):
                name = OPTIONAL_ARG_TEMPLATE.format(name)

            conf[name] = val

        return conf

    def __getstate__(self):
        state = super().__getstate__()
        state["signature"] = Signature([p.replace(annotation=deconstruct_generic(p.annotation))
                                        for p in self.signature.parameters.values()])
        return state

    def __setstate__(self, state):
        sig = state.pop("signature")
        state["signature"] = Signature([p.replace(annotation=reconstruct_generic(p.annotation))
                                        for p in sig.parameters.values()])
        self.__dict__.update(state)



# coding:utf-8

from .io import load_config, dump_config, register_dump_config, register_load_config
from .io import allow_unsafe_yaml, require_safe_yaml, ConfigFormat, LEGAL_CONFIG_EXTENSIONS
from .python import is_json_serializable, UnsafePythonSourceConfig, PythonConfigNotSerializable
from .ini import INIConfigUnserializable
